# (c) Deductive 2012-2020, all rights reserved
# This code is licensed under MIT license (see license.txt for details)
import os
import shutil
import glob
from unittest import TestCase
from newtools.optional_imports import pandas as pd
from newtools.optional_imports import s3fs
from newtools import optional_imports, DoggoFileSystem, S3Location




class BaseTest(TestCase):
    """
    A basic test class that compares the existence of files in different locations
    """
    create_test_data = False
    test_dir = os.path.join(os.path.dirname(__file__), "test_data")
    tempdir = os.path.join(test_dir, "temp")

    def setUp(self):
        try:
            pd.set_option("display.max_rows", 100, "display.max_columns", 100)
        except ImportError:
            pass

        try:
            if not self.tempdir.startswith("s3://"):
                os.makedirs(self.tempdir)
        except FileExistsError:
            self.tearDown()
            self.setUp()

    def tearDown(self):
        if not self.tempdir.startswith("s3://"):
            shutil.rmtree(self.tempdir, )

    def compare_file(self, actual, expected):
        print(actual)
        if self.create_test_data:
            shutil.copyfile(actual, expected)

        with open(expected) as expected_file:
            with open(actual, "r") as actual_file:
                self.assertEqual(expected_file.readlines(),
                                 actual_file.readlines(),
                                 "actual file(right) is not as expected (left)")

    def compare_files(self, expected_path):
        for file in glob.glob(os.path.join(self.tempdir, "*")):

            if file.endswith(".gz"):
                old_file = file
                file = file.replace(".gz", "")
                import gzip
                with gzip.open(old_file, "rb") as f_in:
                    with open(file, "wb") as f_out:
                        f_out.write(f_in.read())

            self.compare_file(file, os.path.join(expected_path, os.path.split(file)[1]))


class TestBase(BaseTest):
    """
    Get full coverage by testing the error cases on the optional imports. The actual tests themselves wouldn't run
    if these didn't work

    """

    def test_import_retry(self):
        @optional_imports.AWSRetry.backoff()
        def banana():
            pass

        banana()

    def test_dummy_awsretry(self):
        @optional_imports.DummyAWSRetry.backoff()
        def apple():
            pass

        apple()

    def test_never_imports(self):
        optional_imports.NEVER_IMPORT = True

        my_s3fs = optional_imports.get_s3fs()
        with self.assertRaisesRegex(ImportError, "s3fs is not installed. Failed getting S3FileSystem()"):
            my_s3fs.S3FileSystem.read_timeout = 600

        with self.assertRaisesRegex(ImportError, "s3fs is not installed. Failed calling s3fs()"):
            my_s3fs()(1, 2, 3)

        optional_imports.NEVER_IMPORT = False

        self.assertEqual(optional_imports.get_s3fs(), s3fs)



# (c) 2012-2020 Deductive, all rights reserved
# -----------------------------------------
#  This code is licensed under MIT license (see license.txt for details)

"""

Cross-platform supports for reading, writing, copying, deleting, etc... files from S3 or local

.. parsed-literal::

             ,--._______,-.
           ,','  ,    .  ,_`-.
          / /  ,' , _` ``. |  )       `-..
         (,';'""`/ '"`-._ ` `/ ______    \\
           : ,o.-`- ,o.  )`` -'      `---.))
           : , d8b ^-.   '|   `.      `    `.
           |/ __:_     `. |  ,  `       `    \
           | ( ,-.`-.    ;'  ;   `       :    ;
           | |  ,   `.      /     ;      :    \
           ;-'`:::._,`.__),'             :     ;
          / ,  `-   `--                  ;     |
         /  `                   `       ,      |
        (    `     :              :    ,`      |
         `   `.    :     :        :  ,'  `    :
          `    `|-- `     ` ,'    ,-'     :-.-';
          :     |`--.______;     |        :    :
           :    /           |    |         |   \
           |    ;           ;    ;        /     ;
         _/--' |   -hrr-   :`-- /         `_:_:_|
       ,',','  |           |___ \
       `^._,--'           / , , .)
                          `-._,-'
"""

import shutil
import os
from glob import glob
import re
from newtools.optional_imports import s3fs
from newtools.aws import S3Location
from .doggo import FileDoggo


class DoggoFileSystem:
    """
    Implements common file operations using either S3 or local file system depending on whether the path
    begins "s3://"

    """
    __s3fs = None

    def __init__(self, session=None):
        s3fs.S3FileSystem.read_timeout = 600
        self.__s3fs = s3fs.S3FileSystem(session=session)

    @property
    def _s3fs(self):
        """
        S3FS caching does not respect other applications updating S3 so therefore we invalidate
        the cache before using

        :return: the S3FS File system
        """
        self.__s3fs.invalidate_cache()
        return self.__s3fs

    def is_s3(self, path1, path2=None):
        """
        Returns true if the passed path is on S3

        :param path1: the first path to check
        :param path2: the second path to check
        :raises NotImplementedError: if only one of the two paths in on S3
        :return: True if both are S3, False if neither are, and raises an exception for mixed types
        """
        p1_is_s3 = path1.startswith("s3://")
        p2_is_s3 = p1_is_s3 if path2 is None else path2.startswith("s3://")

        if path2 is not None:
            return p1_is_s3 and p2_is_s3
        else:
            return p1_is_s3

    def _check_folders(self, path):
        if not self.is_s3(path):
            os.makedirs(os.path.dirname(path), exist_ok=True)

    def cp(self, source, destination):
        """
        Copies a file or folder, per shutil copy() or shutil.copytree() depending on
        if source is a /folder/ or a /file.extension

        :param source: source path, folders must be specified by trailing '/'
        :param destination: destination path, folders must be specified by trailing '/'
        """

        if not self.exists(source):
            raise FileNotFoundError('The specified source file/folder cannot be located')

        is_directory = source.endswith('/')

        if self.is_s3(source, destination):
            return self._s3fs.cp(source, destination)

        elif not self.is_s3(source) and self.is_s3(destination):
            return self._s3fs.put(lpath=source,
                                  rpath=destination,
                                  recursive=True)

        elif self.is_s3(source) and not self.is_s3(destination):
            return self._s3fs.get(lpath=destination,
                                  rpath=source,
                                  recursive=is_directory)

        else:
            if is_directory:
                return shutil.copytree(source, destination)
            else:
                self._check_folders(destination)
                return shutil.copy(source, destination)

    def mv(self, source, destination):
        """
        Moves a file per shutil.move() except that it does not copy WITHIN.
        i.e. /location/folderA/ >> /destination/folderA/
        rather than
        i.e. /location/folderA/ >> /destination/folderA/folderA/

        :param source: source path, folders must be specified by trailing '/'
        :param destination: destination path, folders must be specified by trailing '/'
        """

        if not self.exists(source):
            raise FileNotFoundError('The specified source file/folder cannot be located')

        is_directory = source.endswith('/')

        if self.is_s3(source, destination):
            return self._s3fs.mv(source, destination)

        elif not self.is_s3(source) and self.is_s3(destination):
            self._s3fs.put(lpath=source,
                           rpath=destination,
                           recursive=True)
            if is_directory:
                return shutil.rmtree(source)
            else:
                return os.remove(source)

        elif self.is_s3(source) and not self.is_s3(destination):
            self._s3fs.get(lpath=destination,
                           rpath=source,
                           recursive=is_directory)
            return self._s3fs.rm(source, recursive=is_directory)

        else:
            if is_directory:
                shutil.copytree(source, destination)
                return shutil.rmtree(source)

            else:
                self._check_folders(destination)
                shutil.copy(source, destination)
                return os.remove(source)

    def exists(self, path):
        """
        Returns true if a path exists, per os.path.exists()

        :param path: the path to check
        :return: True if the path exists, otherwise False
        """
        if self.is_s3(path):
            return self._s3fs.exists(path)
        else:
            return os.path.exists(path)

    def size(self, path):
        """
        Returns the size of a file per os.path.getsize()

        :param path: the path to check
        :return: the size of the file at this path
        """
        if self.is_s3(path):
            return self._s3fs.size(path)
        else:
            return os.path.getsize(path)

    def rm(self, path, **kwargs):
        """
        Removes a file, per os.remove()

        :param path: the file to remove
        """
        if self.is_s3(path):
            return self._s3fs.rm(path, **kwargs)
        else:
            return os.remove(path)

    def glob(self, glob_string):
        """
        Searched for a file per glob.glob(recursive=True)


        :param glob_string: the path to search
        :return:
        """
        if self.is_s3(glob_string):
            return [S3Location(a) for a in self._s3fs.glob(glob_string)]
        else:
            return glob(glob_string, recursive=True)

    def open(self, path, mode, *args, **kwargs):
        """
        Opens a file, per open()

        :param path: the path to open
        :param mode: the mode to open in
        :param args: any arguments in the FileDoggo class
        :param kwargs: any keyword arguments for the FileDoggo class
        :return: a file handle
        """
        if "w" in mode:
            self._check_folders(path)
        return FileDoggo(path, mode, *args, **kwargs)

    def join(self, path, *paths):
        """
        Joins to paths per os.path.join()
        :param path: the first path
        :param paths: the paths to joins
        :return:
        """
        if self.is_s3(path):
            return S3Location(path).join(*paths)
        else:
            return os.path.join(path, *paths)

    def split(self, path):
        """
        Splits a path into prefix and file, per os.path.split()
        :param path:
        :return:
        """
        if self.is_s3(path):
            loc = S3Location(path)
            if loc.prefix is not None:
                return S3Location(loc.bucket).join(loc.prefix), loc.file
            else:
                return S3Location(loc.bucket), loc.file
        else:
            return os.path.split(path)



import unittest
from newtools.aws import AthenaPartition, S3List
from newtools.db import AthenaClient


class TestLoadPartitions(unittest.TestCase):
    bucket = "newtools-tests-data"
    tablename = "load_partition_test_data"
    region = "eu-west-1"
    prefix = "load_partition_test_data/"
    database = "test_database"
    output_location = "s3://aws-athena-query-results-933373196108-eu-west-1/"
    source_file = "s3://newtools-tests-data/load_partition_test_data/"
    athena_client = None

    @classmethod
    def setUpClass(cls):
        cls.s3_list_folders = S3List(cls.bucket)
        cls.load_partitions = AthenaPartition(cls.bucket)
        cls.athena_client = AthenaClient(region=cls.region, db=cls.database, max_retries=2)

        cls.athena_client.add_query(
            f"""
            CREATE EXTERNAL TABLE IF NOT EXISTS {cls.tablename}(
                col1 string,
                col2 string,
                col3 string)
            PARTITIONED BY (year string, month string)
            ROW FORMAT SERDE
                'org.apache.hadoop.hive.serde2.OpenCSVSerde'
            STORED AS INPUTFORMAT
                'org.apache.hadoop.mapred.TextInputFormat'
            OUTPUTFORMAT
                'org.apache.hadoop.hive.ql.io.HiveIgnoreKeyTextOutputFormat'
            LOCATION
                '{cls.source_file}'
            TBLPROPERTIES ('skip.header.line.count'='1')
            ;""", output_location=cls.output_location)

        cls.athena_client.add_query(
            f"""
            ALTER TABLE {cls.tablename} ADD IF NOT EXISTS 
                PARTITION (year='2019', month='02') LOCATION '{cls.source_file}year=2019/month=02/'
                PARTITION (year='2019', month='0:2') LOCATION '{cls.source_file}year=2019/month=0:2/'
                PARTITION (year='2020', month='01') LOCATION '{cls.source_file}year=2020/month=01/'

            ;""", output_location=cls.output_location)

        cls.athena_client.wait_for_completion()

    @classmethod
    def tearDownClass(cls):
        cls.athena_client.add_query(f"DROP TABLE {cls.tablename}", output_location=cls.output_location)
        cls.athena_client.wait_for_completion()

    def test_list_folder(self):
        folders = [{'Prefix': 'load_partition_test_data/year=2019/'}, {'Prefix': 'load_partition_test_data/year=2020/'}]
        test_folders = self.s3_list_folders.list_folders(self.prefix)
        self.assertListEqual(folders, test_folders)

    def test_list_partitions(self):
        partitions_list = [{'Prefix': 'load_partition_test_data/year=2019/month=02/some_more_directory/'},
                           {'Prefix': 'load_partition_test_data/year=2019/month=0:2/some_more_directory/'},
                           {'Prefix': 'load_partition_test_data/year=2020/month=01/some_more_directory/'},
                           {'Prefix': 'load_partition_test_data/year=2020/month=01/some_more_directory_1/'},
                           {'Prefix': 'load_partition_test_data/year=2020/month=02/some_more_directory/'}]
        test_partitions_list = self.load_partitions.list_partitions('load_partition_test_data')
        self.assertListEqual(partitions_list, test_partitions_list)

    def test_list_partitions_error(self):
        with self.assertRaises(ValueError):
            self.load_partitions.list_partitions()

    def test_get_sql(self):
        sql_query_list = [
            "ALTER TABLE load_partition_test_data ADD IF NOT EXISTS PARTITION(year='2019',month='02') LOCATION 's3://newtools-tests-data/load_partition_test_data/year=2019/month=02/some_more_directory/' PARTITION(year='2019',month='0:2') LOCATION 's3://newtools-tests-data/load_partition_test_data/year=2019/month=0:2/some_more_directory/' PARTITION(year='2020',month='01') LOCATION 's3://newtools-tests-data/load_partition_test_data/year=2020/month=01/some_more_directory/' PARTITION(year='2020',month='01') LOCATION 's3://newtools-tests-data/load_partition_test_data/year=2020/month=01/some_more_directory_1/' PARTITION(year='2020',month='02') LOCATION 's3://newtools-tests-data/load_partition_test_data/year=2020/month=02/some_more_directory/'"]
        test_sql_query_list = self.load_partitions.get_sql('load_partition_test_data')
        self.assertListEqual(sql_query_list, test_sql_query_list)

    def test_generate_sql(self):
        partitions_list = [{'Prefix': 'load_partition_test_data/year=2019/month=02/some_more_directory/'},
                           {'Prefix': 'load_partition_test_data/year=2019/month=0:2/some_more_directory/'},
                           {'Prefix': 'load_partition_test_data/year=2020/month=01/some_more_directory/'},

                           ]

        sql_query_list = [
            "ALTER TABLE load_partition_test_data ADD IF NOT EXISTS PARTITION(year='2019',month='02') LOCATION 's3://newtools-tests-data/load_partition_test_data/year=2019/month=02/some_more_directory/' PARTITION(year='2019',month='0:2') LOCATION 's3://newtools-tests-data/load_partition_test_data/year=2019/month=0:2/some_more_directory/' PARTITION(year='2020',month='01') LOCATION 's3://newtools-tests-data/load_partition_test_data/year=2020/month=01/some_more_directory/'"]

        sql_query_list_test = self.load_partitions.generate_sql('load_partition_test_data', partitions_list)
        self.assertListEqual(sql_query_list, sql_query_list_test)

    def test_generate_sql_with_client(self):
        partitions_list = [{'Prefix': 'load_partition_test_data/year=2019/month=02/'},
                           {'Prefix': 'load_partition_test_data/year=2019/month=0:2/'},
                           {'Prefix': 'load_partition_test_data/year=2020/month=01/'},
                           {'Prefix': 'load_partition_test_data/year=2020/month=02/'}]

        sql_query_list = [
            "ALTER TABLE load_partition_test_data ADD IF NOT EXISTS PARTITION(year='2020',month='02') LOCATION 's3://newtools-tests-data/load_partition_test_data/year=2020/month=02/'"]

        sql_query_list_test = self.load_partitions.generate_sql('load_partition_test_data', partitions_list,
                                                                athena_client=self.athena_client,
                                                                output_location=self.output_location)
        self.assertListEqual(sql_query_list, sql_query_list_test)

    def test_generate_sql_with_client_error(self):
        partitions_list = [{'Prefix': 'load_partition_test_data/year=2019/month=02/'},
                           {'Prefix': 'load_partition_test_data/year=2019/month=0:2/'},
                           {'Prefix': 'load_partition_test_data/year=2020/month=01/'}]

        with self.assertRaises(ValueError):
            self.load_partitions.generate_sql('load_partition_test_data', partitions_list,
                                              athena_client=self.athena_client)



# (c) Deductive 2012-2020, all rights reserved
# This code is licensed under MIT license (see license.txt for details)

from functools import wraps

# global used for testing results without imports
NEVER_IMPORT = False


class DummyClass:
    """
    A dummy class that raises an import error when you try to do anything with it.

    Is returned in place of the requested module/package/function if an Import error occurs
    """

    def __init__(self, name):
        self.name = name

    def __getattr__(self, name):
        raise ImportError('{0} is not installed. Failed getting {1}()'.format(self.name, name))

    def __call__(self, *args, **kwargs):
        raise ImportError('{0} is not installed. Failed calling {0}()'.format(self.name))


class DummyAWSRetry:
    """
    Dummy implementation of AWSRetry class with fake backoff method
    """

    @classmethod
    def backoff(cls, *_, **__):
        def decorator(f):
            @wraps(f)
            def wrapper(*args, **kwargs):
                return f(*args, **kwargs)

            return wrapper

        return decorator


def import_wrapper(default):
    """
    A decorator to return a Dummy function if an ImportError is raised

    :param default: the dummy function to return
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                if NEVER_IMPORT:
                    raise ImportError
                return func(*args, **kwargs)
            except ImportError:
                return default

        return wrapper

    return decorator


# Now define each import in turn


# -----------------------------
@import_wrapper(DummyAWSRetry)
def get_aws_retry():
    from awsretry import AWSRetry
    return AWSRetry


AWSRetry = get_aws_retry()


# -----------------------------
@import_wrapper(DummyClass("s3fs"))
def get_s3fs():
    import s3fs
    return s3fs


s3fs = get_s3fs()


# -----------------------------
@import_wrapper(DummyClass("pandas"))
def get_pandas():
    import pandas
    return pandas


pandas = get_pandas()


# -----------------------------
@import_wrapper(DummyClass("numpy"))
def get_numpy():
    import numpy
    return numpy


numpy = get_numpy()


# -----------------------------

@import_wrapper(DummyClass("pyarrow"))
def get_pyarrow():
    import pyarrow
    return pyarrow


pyarrow = get_pyarrow()


# -----------------------------

@import_wrapper(DummyClass("boto3"))
def get_boto3():
    import boto3
    return boto3


boto3 = get_boto3()


# -----------------------------

@import_wrapper(DummyClass("botocore"))
def get_botocore():
    import botocore
    return botocore


botocore = get_botocore()


# -----------------------------

@import_wrapper(DummyClass("chardet.universaldetector"))
def get_universal_detector():
    from chardet.universaldetector import UniversalDetector
    return UniversalDetector


UniversalDetector = get_universal_detector()


# -----------------------------


@import_wrapper(DummyClass("sqlparse"))
def get_sqlparse():
    import sqlparse
    return sqlparse


sqlparse = get_sqlparse()

# -----------------------------

#  Add new optional imports to the bottom of this file together with the dummy class



import unittest
import logging
from io import StringIO, BytesIO
from os import path
import sqlite3
from newtools import SqlClient, S3Location
import os
import re

import boto3
import botocore

class SqlTest(unittest.TestCase):

    def setUp(self):

        self.base_path = "{0}/test_data/sqlclient/".format(
            path.dirname(path.abspath(__file__)))

        # set up the SQL Lite database
        self.connection = sqlite3.connect(":memory:")

        # enable log capture
        self.log_stream = StringIO()
        sql_logger = logging.getLogger("newtools.sql_client")
        sql_logger.setLevel(logging.DEBUG)
        handler = logging.StreamHandler(self.log_stream)
        handler.setFormatter(logging.Formatter('%(message)s'))
        sql_logger.addHandler(handler)

    def test_execute_df(self):
        sql = SqlClient(self.connection)
        sql.execute_query("{0}create_table.sql".format(self.base_path))

        df = sql.execute_query_to_df("SELECT * FROM test")

        self.assertEqual(df.iloc[0, 0], 'bee bee')
        self.assertEqual(df.iloc[0, 1], 374)
        self.assertEqual(df.shape, (1, 2))
        self.assertEqual(self.log_stream.getvalue()[-16:], "1 rows affected\n")

    def test_execute_df_replace(self):
        sql = SqlClient(self.connection)
        sql.execute_query("{0}create_table.sql".format(self.base_path),
                          replace={"bee bee": "boo boo"})

        df = sql.execute_query_to_df("SELECT * FROM test;")

        self.assertEqual(df.iloc[0, 0], 'boo boo')
        self.assertEqual(df.iloc[0, 1], 374)
        self.assertEqual(df.shape, (1, 2))

        df = sql.execute_query_to_df(
            "UPDATE test set name='' WHERE name = 'boo boo';SELECT * FROM test WHERE name ='';")

        self.assertEqual(df.iloc[0, 0], '')
        self.assertEqual(df.iloc[0, 1], 374)
        self.assertEqual(df.shape, (1, 2))
        self.assertEqual(self.log_stream.getvalue()[-16:], "1 rows affected\n")

        df = sql.execute_query_to_df("UPDATE test set name='baa' WHERE name = '';SELECT * FROM test WHERE name ='';")

        self.assertEqual(df.shape, (0, 0))
        self.assertEqual(self.log_stream.getvalue()[-20:], "No results returned\n")

    def test_execute_csv_with_query_logging(self):
        sql = SqlClient(self.connection, log_query_text=True)
        sql.execute_query("{0}create_table.sql".format(self.base_path))

        sql.execute_query_to_csv("SELECT * FROM test", "{0}sql_test_output.csv".format(self.base_path))

        with open("{0}sql_test_output.csv".format(self.base_path)) as file_out:
            with open("{0}sql_test_input.csv".format(self.base_path)) as file_in:
                self.assertEqual(file_out.read(), file_in.read())

        sql.execute_query_to_csv("SELECT * FROM test", "{0}sql_test_output.csv".format(self.base_path), append=True)

        with open("{0}sql_test_output.csv".format(self.base_path)) as file_out:
            with open("{0}sql_test_input_append.csv".format(self.base_path)) as file_in:
                self.assertEqual(file_out.read(), file_in.read())

        sql.execute_query_to_csv("SELECT * FROM test", "{0}sql_test_output.csv".format(self.base_path))

        with open("{0}sql_test_output.csv".format(self.base_path)) as file_out:
            with open("{0}sql_test_input.csv".format(self.base_path)) as file_in:
                self.assertEqual(file_out.read(), file_in.read())

    def test_exception(self):
        sql = SqlClient(self.connection)
        try:
            sql.execute_query("{0}file_not_found.sql".format(self.base_path))
        except OSError:
            self.assertTrue(True)
            return

        self.assertTrue(False)

    def test_log_params(self):

        sql = SqlClient(self.connection)

        sql_log_file = os.path.join(self.base_path, "sql_log.log")

        sql.execute_query(os.path.join(self.base_path, "query_file.sql"),
                          parameters={'f_names': ('John', 'Terry', 'Eric', 'Terry', 'Michael', 'Graham'),
                                      'last_name': 'Chapman'},
                          dry_run=True,
                          archive_query=sql_log_file)

        reference = """SELECT top 10 * from pythons

where first_name in ('John', 'Terry', 'Eric', 'Terry', 'Michael', 'Graham')

or last_name = 'Chapman';
"""
        with open(sql_log_file) as f:
            self.assertTrue(re.match(r'-- Ran query on: \d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}', f.readline()))
            self.assertEqual("-- Parameters: {'f_names': ('John', 'Terry', 'Eric', 'Terry', 'Michael', 'Graham'), 'last_name': 'Chapman'}\n",
                             f.readline())
            self.assertEqual(reference, f.read())

    def test_log_to_s3(self):

        s3_archive = S3Location('newtools-testing/test_sql_archive/sql_log.log')

        sql = SqlClient(self.connection)
        sql.execute_query(os.path.join(self.base_path, "query_file.sql"),
                          parameters={'f_names': ('John', 'Terry', 'Eric', 'Terry', 'Michael', 'Graham'),
                                      'last_name': 'Chapman'},
                          dry_run=True,
                          archive_query=s3_archive)

        reference = b"""SELECT top 10 * from pythons

where first_name in ('John', 'Terry', 'Eric', 'Terry', 'Michael', 'Graham')

or last_name = 'Chapman';
"""

        s3 = boto3.client('s3')

        log = BytesIO()
        s3.download_fileobj(s3_archive.bucket, s3_archive.key, log)

        log.seek(0)
        self.assertTrue(re.match(b'-- Ran query on: \d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}', log.readline()))
        self.assertEqual(
            b"-- Parameters: {'f_names': ('John', 'Terry', 'Eric', 'Terry', 'Michael', 'Graham'), 'last_name': 'Chapman'}\n",
            log.readline())
        self.assertEqual(reference, log.read())

    def test_log_to_s3_fails(self):

        s3_archive = S3Location('notabucket/nopath/sql_log.log')

        sql = SqlClient(self.connection)
        with self.assertRaises(botocore.exceptions.ClientError):
            sql.execute_query(os.path.join(self.base_path, "query_file.sql"),
                              parameters={'f_names': ('John', 'Terry', 'Eric', 'Terry', 'Michael', 'Graham'),
                                          'last_name': 'Chapman'},
                              dry_run=True,
                              archive_query=s3_archive)



from .cached_query import CachedPep249Query, CachedAthenaQuery, BaseCachedQuery
from .sql_client import SqlClient
from .athena import AthenaClient

__all__ = ['CachedAthenaQuery', 'CachedPep249Query', 'BaseCachedQuery', 'SqlClient', 'AthenaClient']


# (c) 2012-2020 Deductive, all rights reserved
# -----------------------------------------
#  This code is licensed under MIT license (see license.txt for details)

from io import BytesIO
import os
from csv import Sniffer, Error
from newtools.optional_imports import UniversalDetector, pandas as pd


from .doggo import PandasDoggo
from .fs import DoggoFileSystem


class CSVDoggo:
    """
    Loads and saves CSV files from pandas DataFrames. Supports GZIP, Parquet, S3, local file system interchangably

    """
    _load_map = {'line_terminator': 'lineterminator',
                 'force_dtype': 'dtype',
                 'nan_values': "na_values"}

    _save_map = {'delimiter': 'sep',
                 'force_dtype': 'dtype',
                 'nan_values': 'na_rep'}

    SAMPLE_SIZE = 1024 * 1024

    def __init__(self, base_path="", detect_parameters=False, *args, **kwargs):
        """

        :param base_path: the base path to use for subsequent save and load operations
        :param args: args to pass to save and load functions
        :param kwargs: kwargs to pass to save and load functions
        """
        self._base_path = base_path
        self._doggo = PandasDoggo()
        self._args = args
        self._kwargs = kwargs

        self.detect_parameters = detect_parameters

    def _get_file(self, file):
        return os.path.join(self._base_path, file)

    @staticmethod
    def _get_encoding(sample):
        detector = UniversalDetector()
        for line in BytesIO(sample).readlines():
            detector.feed(line)
            if detector.done:  # pragma: no cover
                break
        detector.close()
        return detector.result["encoding"]

    def _sniff_parameters(self, file):
        """
        sets instance variable with detected parameters from file
        :param file: name of file to be considered (relative to self.base_path)
        :return: True if able to detect parameters
        """
        # if we are using the default parameters, then attempt to guess them

        # create a sample...
        full_path = self._get_file(file)
        with DoggoFileSystem().open(full_path, mode="rb") as f:
            sample = f.read(self.SAMPLE_SIZE)

        # get the encoding...
        encoding = self._get_encoding(sample)
        if encoding is None:
            encoding = 'windows-1252'

        # now decode the sample
        try:
            sample = sample.decode(encoding)
        except UnicodeDecodeError:
            encoding = "windows-1252"
            try:
                sample = sample.decode(encoding)
            except UnicodeDecodeError as e:
                raise ValueError(str(e))

        # use the sniffer to detect the parameters...
        sniffer = Sniffer()
        dialect = sniffer.sniff(sample)
        delimiter = dialect.delimiter
        # the detector always seems to find a header... below line is in case it does not
        header = 0 if sniffer.has_header(sample) else -1
        quotechar = dialect.quotechar

        return {
            "encoding": encoding,
            "delimiter": delimiter,
            "header": header,
            "quotechar": quotechar
        }

    def _get_args(self, args):
        return args + self._args

    def _get_kwargs(self, kwargs, field_map):
        kw = self._kwargs.copy()
        kw.update(kwargs)
        for key in list(kw.keys()):
            if key in field_map:
                kw[field_map[key]] = kw.pop(key)
            elif key.startswith("csv_"):
                kw[key[4:]] = kw.pop(key)

        if kw.get("header", 0) == -1:
            kw.pop("header")

        if 'na_rep' in kw:
            kw['na_rep'] = kw['na_rep'][0]

        return kw

    def _get_path(self, file):
        return os.path.join(self._base_path, file)

    def load_df(self, file, *args, **kwargs):
        """
        Loads a data frame

        :param file: the file to load
        :param args: any args to use
        :param kwargs: any keyword args to use
        :return: a pandas data frame
        """

        try:
            return self._doggo.load_csv(self._get_path(file),
                                        *self._get_args(args),
                                        **self._get_kwargs(kwargs, self._load_map))
        except (UnicodeError, UnicodeDecodeError, pd.errors.ParserError) as e:
            err_msg = str(e)
            if self.detect_parameters:
                kwargs.update(self._sniff_parameters(file))
                return self._doggo.load_csv(self._get_path(file),
                                            *self._get_args(args),
                                            **self._get_kwargs(kwargs, self._load_map))
            raise ValueError(err_msg)

    def save_df(self, df, file, *args, **kwargs):
        """
        Saves a data frame

        :param df: the data frame to save
        :param file: the path to save to
        :param args: any args to use
        :param kwargs: any keyword args to use
        """
        if 'index' not in kwargs:
            kwargs['index'] = False
        return self._doggo.save_csv(df, self._get_path(file), *self._get_args(args),
                                    **self._get_kwargs(kwargs, self._save_map))

    def df_to_string(self, df, *args, **kwargs):
        """
        Returns a formatted string from a dataframe using the specified
        configuration for the class

        :param df: the data frame to cast to string
        """

        return df.to_csv(*self._get_args(args), **self._get_kwargs(kwargs, self._save_map))



# (c) 2012-2018 Deductive, all rights reserved
# -----------------------------------------
#  This code is licensed under MIT license (see license.txt for details)
import queue
import time
import datetime
import logging


class Task:
    """
    An abstraction representing a Single task
    """

    def __init__(self, name, arguments):
        self.is_complete = False
        self.error = None
        self.arguments = arguments
        self.id = None
        self.retries = 0
        self.name = name


class TaskQueue:
    """
    This is an abstract class that contains all required common functionality to
    support implementation of queues in aws client libraries.

    Two separate queues are maintained:

    active_queue
    ------------

    - Only tasks in this queue are allowed to run.
    - Once tasks are completed they are removed from this queue.
    - No of tasks in active queue <= max_size.

    pending_queue
    -------------

    - Contains tasks that are awaiting execution.
    - Tasks from pending_queue are added to active_queue in FIFO fashion.
    """

    def __init__(self, max_size,
                 retry_limit=3,
                 logger=logging.getLogger("newtools.queue")):
        self.pending_tasks = queue.Queue()
        self.active_queue = []
        self.max_size = max_size
        self.retry_limit = retry_limit
        self._time_first_job_added = datetime.datetime.now()
        self._tasks_completed = 0
        self._logger = logger

    def add_task(self, name, args):
        """This method adds a tasks to the pending_tasks queue"""
        task = Task(name, args)
        self.pending_tasks.put(task)
        self._empty_active_queue()
        self._fill_active_queue()
        return task

    def _empty_active_queue(self):
        """
        Removes completed task from active queue and populates freed spots with tasks
        in the front of pending_tasks queue
        """
        # Remove completed tasks from active queue
        for index, task in enumerate(self.active_queue):
            self._update_task_status(task)
            if task.error:
                if task.retries < self.retry_limit:
                    self._logger.info("Retrying job {0}, previously raised error with error {1}".
                                      format(task.name, task.error))
                    task.retries += 1
                    task.error = None
                    self._trigger_task(task)
                else:
                    self._tasks_completed = self._tasks_completed + 1
                    task.is_complete = True
                    self.active_queue.pop(index)
                    raise TimeoutError("The task exceeded the retry limit: {0} [name: '{1}', id: {2}]".format(
                        task.error, task.name, task.id))

            if task.is_complete:
                self._tasks_completed = self._tasks_completed + 1
                self._logger.debug("Task is complete: ID {0}".format(task.id))
                self.active_queue.pop(index)

    @property
    def number_active(self):
        return len(self.active_queue)

    @property
    def number_pending(self):
        return self.pending_tasks.qsize()

    def _fill_active_queue(self):
        # Add add tasks to active queue if size of queue is less the max query limit
        for i in range(0, self.max_size - self.number_active):
            if not (self.pending_tasks.empty()):
                task = self.pending_tasks.get()
                self.active_queue.append(task)
                self._trigger_task(task)
            else:
                break

    @property
    def time_per_task(self):
        return (datetime.datetime.now() - self._time_first_job_added) / self._tasks_completed

    @property
    def time_remaining(self):
        if self._tasks_completed > 0:
            return self.time_per_task * (self.number_active + self.number_pending)
        else:
            return "Unknown"

    def wait_for_completion(self):
        """
        This method runs until execution of all tasks is completed
        """
        while len(self.active_queue):
            logging.info("{} tasks are awaiting exection".format(self.number_active))
            self._empty_active_queue()
            self._fill_active_queue()
            time.sleep(10)

        self._tasks_completed = 0
        self._time_first_job_added = datetime.datetime.now()

    def _trigger_task(self, task):  # pragma: nocover
        """
           This function should implement functionality to start task
           Once triggered is_active should be set to True
        """
        raise NotImplementedError("must be implemented by subclass")

    def _update_task_status(self, task):  # pragma: nocover
        """Updates task status and returns updated task object
           Also handles retries
        """
        raise NotImplementedError("must be implemented by subclass")

    def _empty_pending_queue(self):
        """
        Empty pending queue of tasks - prevent them from being run
        :return: None
        """
        while not self.pending_tasks.empty():
            self.pending_tasks.get()



import unittest
import warnings
import os
import tempfile
import shutil

from newtools.optional_imports import boto3
from newtools.optional_imports import pandas as pd

from newtools import CSVDoggo, S3Location, log_to_stdout

from s3fs import S3FileSystem

FpCSVEncodingError = ValueError

logger = log_to_stdout("newtools")

s3_bucket = "newtools-testing"


class CSVFileTests(unittest.TestCase):
    saved_suffix = ".saved"
    root_prefix = "csvhandler"
    s3 = None
    local_base_path = None

    @classmethod
    def setUpClass(cls):

        # Original test data location
        cls.orig_base_path = "{0}/test_data/{1}/".format(
            os.path.dirname(os.path.abspath(__file__)), cls.root_prefix)

        # Upload test data to s3
        cls.s3_base_path = f"s3://{s3_bucket}/{cls.root_prefix}/"
        cls.s3 = S3FileSystem()
        for file in cls.s3.glob(f"{cls.s3_base_path}*"):
            cls.s3.rm(file)

        cls.s3.put(lpath=cls.orig_base_path,
                   rpath=cls.s3_base_path,
                   recursive=True)

        # Save test data to local folder
        cls.local_base_path = tempfile.mkdtemp() + '/'
        os.mkdir(cls.local_base_path + cls.root_prefix)

    @classmethod
    def tearDownClass(cls):
        # Delete uploaded/created files
        for file in cls.s3.glob(f"{cls.s3_base_path}*"):
            cls.s3.rm(file)
        shutil.rmtree(cls.local_base_path, ignore_errors=True)

    def test_files(self):

        warnings.simplefilter("ignore", ResourceWarning)

        # Test file location
        os.makedirs(self.local_base_path + self.root_prefix, exist_ok=True)

        # Get list of the files
        files = [S3Location(f).file for f in self.s3.glob(f"s3://{s3_bucket}/{self.root_prefix}/*.csv") if
                 'unable_to_decode' not in f]

        # Create the handlers
        csv_orig = CSVDoggo(base_path=self.orig_base_path,
                            detect_parameters=True)
        csv_s3 = CSVDoggo(base_path=self.s3_base_path,
                          detect_parameters=True)
        csv_local = CSVDoggo(base_path=self.local_base_path,
                             detect_parameters=True)
        csv_empty = CSVDoggo(detect_parameters=True)

        # Alternate how the path is split
        use_empty_base = True

        for file in files:

            logger.info("testing file {0}".format(file))

            # Read file from S3
            if use_empty_base:
                if file.endswith(".gzip"):
                    df1 = csv_empty.load_df(self.s3_base_path + file, compression="gzip")
                else:
                    df1 = csv_empty.load_df(self.s3_base_path + file)
            else:
                if file.endswith(".gzip"):
                    df1 = csv_s3.load_df(file, compression="gzip")
                else:
                    df1 = csv_s3.load_df(file)

            # Write file to S3
            if use_empty_base:
                csv_empty.save_df(df1, self.s3_base_path +
                                  file + self.saved_suffix)
            else:
                csv_s3.save_df(df1, file + self.saved_suffix)

            # Check DataFrames match
            if use_empty_base:
                df2 = csv_empty.load_df(
                    self.s3_base_path + file + self.saved_suffix)
            else:
                df2 = csv_s3.load_df(file + self.saved_suffix)
            pd.testing.assert_frame_equal(df1, df2)

            # Check strings match
            str1 = csv_s3.df_to_string(df1)
            str2 = csv_s3.df_to_string(df2)
            self.assertMultiLineEqual(str1, str2)

            # Read file locally
            if use_empty_base:
                df3 = csv_empty.load_df(self.orig_base_path + file)
            else:
                df3 = csv_orig.load_df(file)

            # Write file locally
            if use_empty_base:
                csv_empty.save_df(df3, self.local_base_path +
                                  file + self.saved_suffix)
            else:
                csv_local.save_df(df3, file + self.saved_suffix)

            # Check DataFrames match
            if use_empty_base:
                df4 = csv_empty.load_df(
                    self.local_base_path + file + self.saved_suffix)
            else:
                df4 = csv_local.load_df(file + self.saved_suffix)
            pd.testing.assert_frame_equal(df3, df4)

            # Check strings match
            str3 = csv_orig.df_to_string(df3)
            str4 = csv_local.df_to_string(df4)
            self.assertMultiLineEqual(str3, str4)

            use_empty_base = not use_empty_base


class CSVOtherTests(unittest.TestCase):
    local_base_path = None

    @classmethod
    def setUpClass(cls):
        # Original test data location
        cls.orig_base_path = "{0}/test_data/csvhandler/".format(
            os.path.dirname(os.path.abspath(__file__)))

        cls.local_base_path = tempfile.mkdtemp() + '/'
        cls.s3_base_path = "s3://{}/".format(s3_bucket)

        cls.csv = CSVDoggo(base_path=cls.orig_base_path)
        cls.csv_windows = CSVDoggo(base_path=cls.orig_base_path, csv_encoding="Windows-1252")

    @classmethod
    def tearDownClass(cls):
        # Delete uploaded/created files
        shutil.rmtree(cls.local_base_path, ignore_errors=True)

    def test_encoding_error(self):
        with self.assertRaises(FpCSVEncodingError):
            csv = CSVDoggo(base_path=self.orig_base_path)
            csv.load_df("test_cities_reference1252.csv")

    def test_csv_parameter(self):
        csv = CSVDoggo(base_path=self.orig_base_path)
        csv.load_df("test_cities_reference1252.csv",
                    csv_encoding="Windows-1252")

    def test_gzip(self):
        csv = CSVDoggo(base_path=self.orig_base_path,
                       compression="gzip")
        df1 = csv.load_df("test.csv.gz")
        csv.save_df(df1, "test2.csv.gz")
        df2 = csv.load_df("test2.csv.gz")
        pd.testing.assert_frame_equal(df1, df2)

    def test_zip(self):
        with self.assertRaises(NotImplementedError):
            csv = CSVDoggo(base_path=self.orig_base_path, compression="zip")
            df1 = csv.load_df("test.csv.zip")
            csv.save_df(df1, "test2.csv.zip")
            df2 = csv.load_df("test2.csv.zip")
            pd.testing.assert_frame_equal(df1, df2)

    def test_forcing_dtype(self):
        csv = CSVDoggo(base_path=self.orig_base_path)
        df1 = csv.load_df("test_int_is_unique_clean.csv", force_dtype=str)
        self.assertEqual(str(df1['TotalEpisodes'].dtype), 'object')

    def test_no_header(self):
        csv = CSVDoggo(base_path=self.orig_base_path)
        df = csv.load_df("test_cities_reference1252.csv",
                         csv_encoding="Windows-1252",
                         header=-1,
                         force_dtype=str)

        csv2 = CSVDoggo(base_path=self.local_base_path,
                        header=-1)
        csv2.save_df(df, "banana.csv")
        df2 = csv2.load_df("banana.csv",
                           force_dtype=str)

        pd.testing.assert_frame_equal(df, df2, check_index_type=False)

    def test_sniffer(self):
        csv = CSVDoggo(base_path=self.orig_base_path)
        csv._sniff_parameters("email_test.csv")

    def test_sniffer_unable_to_decode(self):
        """
        cannot find encoding for encrypted file - need to decrypt first
        """
        with self.assertRaises(FpCSVEncodingError):
            self.csv._sniff_parameters("unable_to_decode.csv")

    def test_nan_values(self):
        """
        the test data should actually include nans to make this test better!
        """
        csv_nan = CSVDoggo(base_path=self.orig_base_path, nan_values=["1", "2"])
        temp_df = csv_nan.load_df("email_test.csv")
        csv_nan.save_df(temp_df, "temporary.csv")
        os.remove(os.path.join(self.orig_base_path, "temporary.csv"))

    def test_pd_kwargs_for_reading(self):
        """
        allows pass through of features not directly supported by CSVDoggo
        (test by making sure that blank entries are read in as blanks)
        """
        CSVDoggo(base_path=self.orig_base_path,
                 pd_kwargs=dict(keep_na_values=False, ))

    def test_basic_chunksize(self):
        with self.assertRaises(NotImplementedError):
            csv = CSVDoggo(base_path=self.orig_base_path)
            dfs = csv.load_df("test_cities_reference1252.csv",
                              csv_encoding="Windows-1252",
                              chunksize=10)

            self.assertEqual([len(df) for df in dfs], [10, 10, 10, 10, 9])

    def test_s3_chunks(self):
        with self.assertRaises(NotImplementedError):
            s3c = boto3.client("s3")
            csvh = CSVDoggo(base_path=self.orig_base_path)
            df_lt = csvh.load_df("email_test.csv", chunksize=5)
            df_init = self.csv.load_df("email_test.csv")
            pd.testing.assert_frame_equal(df_init.head(5), df_lt.get_chunk(5))

    def test_s3_nrows(self):
        csv = CSVDoggo(base_path=self.orig_base_path)
        df = csv.load_df("test_cities_reference1252.csv",
                         csv_encoding="Windows-1252",
                         nrows=10)

        self.assertEqual(len(df), 10)

    def test_usecols(self):
        csv = CSVDoggo(base_path=self.orig_base_path)
        df = csv.load_df("test_cities_dirty.csv", usecols=['city'])
        expected_df = csv.load_df("test_usecols_expected.csv")
        pd.testing.assert_frame_equal(expected_df, df)



import logging
import os
import csv
import datetime
from importlib import import_module
from io import BytesIO

from newtools.aws import S3Location
from newtools.optional_imports import boto3, pandas as pd, sqlparse


def _clean_dict(d):
    block_words = ["secret", "password"]
    clean = dict()
    for param in d:
        if any(x.lower() in param.lower() for x in block_words):
            clean[param] = "*" * len(d[param])
        else:
            clean[param] = d[param]

    return clean


class QueryArchiver:
    _archive = None

    def __init__(self, file):
        self._file = file
        self._archive = BytesIO()

    @property
    def is_s3(self):
        return self._file.startswith('s3://')

    @property
    def am_archiving(self):
        return self._file is not None

    def __enter__(self):
        if self.am_archiving and not self.is_s3:
            os.makedirs(os.path.dirname(self._file), exist_ok=True)

            # clear the file
            open(self._file, 'w').close()

        return self

    def log(self, logged_query, parameters):
        if self.am_archiving:
            self._archive.write(
                '-- Ran query on: {:%Y-%m-%d %H:%M:%S}\n'.format(datetime.datetime.now()).encode('utf-8'))
            self._archive.write('-- Parameters: {0}\n'.format(_clean_dict(parameters)).encode('utf-8'))
            self._archive.write(logged_query.encode('utf-8') + b';\n')

    def __exit__(self, exc_type, exc_val, exc_tb):

        if self.am_archiving:
            self._archive.seek(0)
            if self.is_s3:
                s3_path = S3Location(self._file)
                s3 = boto3.client('s3')
                _ = self._archive.read()  # boto closes this, but i want to keep it just in case
                self._archive.seek(0)
                s3.upload_fileobj(self._archive, s3_path.bucket, s3_path.key)
            else:
                with open(self._file, 'wb') as f:
                    f.write(self._archive.read())


class ParamParser:
    """
    Class to parse SQL parameters into a common format
    """

    named_styles = {
        'named': ':{name}',
        'pyformat': '%({name}){type}'
    }

    sequence_styles = {
        'qmark': '?',
        'numeric': ':{index}',
        'format': '%{type}',
    }

    styles = {**named_styles, **sequence_styles}

    index = 0

    def _get_param(self, param):
        self.index = self.index + 1
        return self.styles[self.paramstyle].format(name=param, type="s", sequence=self.index)

    def __init__(self, paramstyle, logger):
        self.paramstyle = paramstyle
        logger.debug("Queries will use paramstyle = {0}".format(paramstyle))
        if paramstyle not in ('pyformat', 'qmark'):  # pragma: nocover
            logger.warning("The paramstyle {0} has not been tested and may not work".format(paramstyle))

    @staticmethod
    def parse_for_logging(sql, params):

        return sql.format(**_clean_dict(params))

    def parse_sql_params(self, sql, params):

        # validate
        self.parse_for_logging(sql, params)
        if len(params) > 0:
            self.index = 0
            new_sql = ""
            new_params = []
            for segment in sql.replace("'{", "{").replace("}'", "}").replace("{", "}").split("}"):
                if segment in params:
                    new_sql = new_sql + self._get_param(segment)
                    new_params.append(params[segment])
                else:
                    new_sql = new_sql + segment

            if self.paramstyle in self.named_styles:  # pragma: nocover
                # this is not tested because we only have sql lite client for testing which does not support
                # named params.
                return new_sql, params
            else:
                return new_sql, new_params
        else:
            return sql, params


class SqlClient:
    """
    A wrapper for PEP249 connection objects to provide additional logging and simple execution
    of queries and optional writing out of results to DataFrames or CSV

    The client runs multi-statement SQL queries from file or from strings and can return the
    result of the final SQL statement in either a DataFrame or as a CSV.

    Archives the text of the executed queries to an optionally specified location.

    Parameters:
    - db_connection - a connection object from a PEP249 compliant class
    """

    def __init__(self,
                 db_connection,
                 logging_level=logging.DEBUG,
                 log_query_text=False,
                 logger=logging.getLogger("newtools.sql_client")):

        self._logger = logger
        self.connection = db_connection
        self.cursor = self.connection.cursor()
        paramstyle = import_module(db_connection.__class__.__module__.split(".")[0]).paramstyle
        self.parser = ParamParser(paramstyle, self._logger)
        self.logging_level = logging_level
        self._counter = 0
        self._log_query_text = log_query_text

    def _log_query(self, query, parameters):
        self._logger.log(self.logging_level, "Executing part {0}: {1}".format(self._counter, query.split("\n")[0]))
        self._logger.debug("Parameters {0}".format(_clean_dict(parameters)))

        if self._log_query_text:
            for line in query.split("\n"):
                self._logger.log(self.logging_level, "{0:03} {1}".format(self._counter, line))
                self._increment_counter()

    def _report_rowcount(self, execution_time):
        if self.cursor.rowcount >= 0:
            self._logger.log(self.logging_level, "Completed in {0}s. {1} rows affected".format(
                execution_time.seconds,
                self.cursor.rowcount))
        else:
            self._logger.log(self.logging_level, "Completed in {0}s".format(execution_time.seconds))

    def _get_queries(self, query_file, replace):
        replace = dict() if replace is None else replace
        if query_file[-4:] == ".sql":
            if os.path.isfile(query_file):
                self._logger.log(self.logging_level, "Loading query from {0}".format(query_file))
                f = open(query_file, "r")
                text = f.read()
                f.close()
            else:
                self._logger.error("File {0} does not exist".format(query_file))
                raise OSError("File {0} does not exist".format(query_file))
        else:
            text = query_file

        for key in replace:
            text = text.replace(key, replace[key])

        # split into multiple commands....
        if ';' in text:
            # use sql parse if installed, otherwise split on every ;
            generator = sqlparse.split(text)

            # split into multiple commands....
            for command in generator:
                if command.strip() not in ["", "';'"]:
                    yield command.strip()
        else:
            yield text.strip()

    def _run_and_log_sql(self, command, parameters, query_archiver, pandas=False, dry_run=False):
        parameters = dict() if parameters is None else parameters

        sql, params = self.parser.parse_sql_params(command, parameters)

        df = None

        if command != '':
            self._log_query(command, parameters)
            query_archiver.log(self.parser.parse_for_logging(command, parameters), parameters)
            start_time = datetime.datetime.now()
            if not dry_run:
                if pandas:
                    df = pd.read_sql(sql,
                                     self.connection,
                                     params=params)
                else:
                    self.cursor.execute(sql, params)

                self._report_rowcount(datetime.datetime.now() - start_time)

        return df

    def _reset_counter(self):
        self._counter = 1

    def _increment_counter(self):
        self._counter = self._counter + 1

    def _execute_queries(self, queries, parameters, first_to_run, query_archiver, dry_run):
        self._reset_counter()
        for command in queries:
            if self._counter >= first_to_run:
                self._run_and_log_sql(command=command,
                                      parameters=parameters,
                                      pandas=False,
                                      query_archiver=query_archiver,
                                      dry_run=dry_run
                                      )
            self._increment_counter()

    def execute_query(self, query, parameters=None, replace=None, first_to_run=1, archive_query=None, dry_run=False):
        """
        Runs a query and ignores any output

        Parameters:
        - query - the query to run, either a SQL file or a SQL query
        - parameters - a dict of parameters to substitute in the query
        - replace - a dict or items to be replaced in the SQL text
        - first_to_run - the index of the first query in a multi-command query to be executed
        - archive_query - save the query that is run to file. Default=False,

        """
        with QueryArchiver(archive_query) as archiver:
            self._execute_queries(queries=self._get_queries(query, replace),
                                  parameters=parameters,
                                  first_to_run=first_to_run,
                                  query_archiver=archiver,
                                  dry_run=dry_run)

        self.connection.commit()

    def execute_query_to_df(self, query, parameters=None, replace=None, first_to_run=1, dry_run=False,
                            archive_query=None):
        """
        Runs a query and returns the output of the final statement in a DataFrame.

        Parameters:
        - query - the query to run, either a SQL file or a SQL query
        - parameters - a dict of parameters to substitute in the query
        - replace - a dict or items to be replaced in the SQL text
        - first_to_run - the index of the first query in a multi-command query to be executed
        - archive_query - save the query that is run to file. Default=False,

        """

        commands = [command for command in self._get_queries(query, replace)]

        with QueryArchiver(archive_query) as archiver:
            self._execute_queries(queries=commands[:-1],
                                  parameters=parameters,
                                  first_to_run=first_to_run,
                                  query_archiver=archiver,
                                  dry_run=dry_run)

            # now run the last one as a select
            df = self._run_and_log_sql(command=commands[-1],
                                       parameters=parameters,
                                       pandas=True,
                                       dry_run=dry_run,
                                       query_archiver=archiver)

        self.connection.commit()

        if not dry_run and len(df) == 0:
            self._logger.info("No results returned")
            return pd.DataFrame()
        else:
            return df

    def execute_query_to_csv(self,
                             query,
                             csvfile,
                             parameters=None,
                             replace=None,
                             append=False,
                             first_to_run=1,
                             archive_query=None,
                             dry_run=False):
        """
        Runs a query and writes the output of the final statement to a CSV file.

        Parameters:
        - query - the query to run, either a SQL file or a SQL query
        - csvfile - the file name to save the query results to
        - parameters - a dict of parameters to substitute in the query
        - replace - a dict or items to be replaced in the SQL text
        - first_to_run - the index of the first query in a multi-command query to be executed
        """
        with QueryArchiver(archive_query) as archiver:
            self._execute_queries(queries=self._get_queries(query, replace),
                                  parameters=parameters,
                                  first_to_run=first_to_run,
                                  query_archiver=archiver,
                                  dry_run=dry_run)

        if not dry_run:
            # delete an existing file if we are not appending
            if os.path.exists(csvfile) and append:
                file_mode = 'a'
            else:
                file_mode = 'w'

            # now get the data
            with open(csvfile, file_mode) as f:
                writer = csv.writer(f, delimiter=',')

                # write the header if we are writing to the beginning of the file
                if file_mode == 'w':
                    writer.writerow([desc[0] for desc in self.cursor.description])

                for row in self.cursor:
                    writer.writerow(row)



from newtools.optional_imports import boto3
from newtools.aws import S3Location
from newtools.doggo.doggo import FileDoggo
import os


class S3List:

    def __init__(self, bucket, s3_client=None):
        """
        :param bucket: The s3 bucket to look for data in
        :param s3_client: optional pass in of a pre created boto3 s3 client
        """
        self.s3_client = s3_client if s3_client else boto3.client('s3')
        self.bucket = bucket

    def list_folders(self, prefix=''):
        """
        :param prefix: prefix the data is stored under if empty string will scan at highest level of bucket
        :return: returns a list of the folders at that level
        """
        result_list = []

        # force prefix end with "/"
        if prefix != '' and not prefix.endswith("/"):
            prefix = prefix + "/"

        paginator = self.s3_client.get_paginator("list_objects_v2")
        pages = paginator.paginate(Bucket=self.bucket,
                                   Delimiter='/',
                                   Prefix=prefix)

        for page in pages:
            if page.get('CommonPrefixes'):
                common_prefixes = page.get('CommonPrefixes')
                result_list.extend(common_prefixes)

        return result_list


class AthenaPartition:
    """
    The AthenaPartition class will recurse to a specified depth in S3 and create SQL to add all partitions
    to a table in Athena. This is significantly faster than using ``MSCK REPAIR TABLE`` as it only searches
    to the specified depth.

    """

    def __init__(self, bucket, s3_client=None):
        """
        :param bucket: S3 bucket where the data is stored
        :param s3_client: optional param of pre initialised s3 client
        """

        self.s3_client = s3_client
        self.s3_bucket = bucket

    def list_partitions(self, table=None, partition_depth=None, s3_path=None):
        """
        :param table: The athena table partitions are being loaded for, either this or s3_path has to be specified
        :param partition_depth: The number of partition levels to look for, optional param should be specified if known significant performance impact
        :param s3_path: The prefix for where the data is stored, can be set to blank string to look at top level
                        of the bucket
        :return: A list of all the full paths of the partitions
        """
        partition_depth = partition_depth or 100

        if s3_path is None:

            if table is None:
                raise ValueError('Either table or s3_path must be set both can not be None')

            s3_path = table

        list_files = S3List(self.s3_bucket, self.s3_client)

        partitions_list = list_files.list_folders(s3_path)
        partitions_tracked = 0

        while partitions_tracked < partition_depth:
            temp_part_list = []
            for partition in partitions_list:
                partitions_path_temp = list_files.list_folders(partition.get('Prefix'))
                """
                For each partition in the partition_list we check if there is any more depth, once we reach filename
                `list_files.list_folders(partition.get('Prefix'))` will return [] we check for this and break from
                checking the rest of the partitions, as all partitions have the same depth.
                """
                if not partitions_path_temp:
                    break
                else:
                    temp_part_list.extend(partitions_path_temp)

            # If temp_part_list is still [] no further partitions have been found so we break out of the while loop.

            if temp_part_list:
                partitions_list = temp_part_list
            else:
                break
            partitions_tracked += 1
        return partitions_list

    def generate_sql(self, table, partitions_list, s3_path=None, athena_client=None, output_location=None,
                     athena_s3_client=None):
        """
                :param table: The athena table partitions are being loaded for
                :param partitions_list: The list of all the partitions to generate the alter table query for
                :param s3_path: The prefix for where the data is stored, optional, can also be empty string to look at
                                the toplevel of the bucket
                :param athena_client: The current Athena Client, optional, will be used to list partitions already loaded
                :param output_location: Output location of queries parsed into athena_client
                :param athena_s3_client: The s3 client that matches with the athena client passed in to be able to
                access the queries produced by it if in different account to default
                :return: A list of queries that can be run to add all partitions
                """
        s3_path = table if s3_path is None else s3_path

        if athena_client is not None:
            if output_location is None:
                raise ValueError('output_location must be set to use athena_client')

            query = athena_client.add_query("SHOW PARTITIONS {0}".format(table),
                                            name="show_partitions_athena",
                                            output_location=output_location)

            athena_client.wait_for_completion()
            fm = FileDoggo(os.path.join(query.arguments["output_location"], "{}.txt".format(query.id)), mode='rb',
                           compression=None, client=athena_s3_client)
            with fm as f:
                athena_list = f.read().decode().splitlines()
            athena_list = [{'Prefix': '{table}/{partition}/'.format(table=s3_path, partition=partition)} for partition in
                           athena_list]
            partitions_list = [i for i in partitions_list if i not in athena_list]

        partition_string_list = []
        for i in partitions_list:
            partition_string = i.get('Prefix')
            location = S3Location(bucket=self.s3_bucket, key=partition_string, ignore_double_slash=True)
            partition_string = partition_string.replace(s3_path, "")
            partition_string = partition_string.rstrip('/')

            partitions_string_list = [
                '{key}=\'{value}\''.format(key=partition.split('=')[0], value=partition.split('=')[1]) for partition in
                partition_string.split('/') if '=' in partition]

            partition_string = ",".join(partitions_string_list)

            partition_string_formatted = ' PARTITION({}) LOCATION \'{}\''.format(partition_string, location)

            partition_string_list.append(partition_string_formatted)

        query_string_base = "ALTER TABLE {table} ADD IF NOT EXISTS".format(table=table)

        full_query_string_list = []
        temp_string_build = ''
        for count, string in enumerate(partition_string_list):
            temp_string_build += string
            if len(temp_string_build.encode('utf-8')) >= 250000 or (count + 1) == len(partition_string_list):
                query_string_build = query_string_base + temp_string_build
                full_query_string_list.append(query_string_build)
                temp_string_build = ''

        return full_query_string_list

    def get_sql(self, table, partition_depth=None, s3_path=None, athena_client=None, output_location=None,
                athena_s3_client=None):
        """
        :param table: The athena table partitions are being loaded for
        :param partition_depth: The number of partition levels to look for, optional param should be specified if
                                known as will yield best performance.
        :param s3_path: The prefix for where the data is stored, optional, can also be empty string to look at the top
                        level of the bucket
        :param athena_client: The current Athena Client, optional, will be used to list partitions already loaded
        :param output_location: Output location of queries parsed into athena_client
        :param athena_s3_client: The s3 client that matches with the athena client passed in to be able to
                access the queries produced by it if in different account to default
        :return: A list of queries that can be run to add all partitions, each query is limited to 10 partitions
        """

        partitions_list = self.list_partitions(table, partition_depth, s3_path)

        return self.generate_sql(table, partitions_list, s3_path, athena_client, output_location, athena_s3_client)



# (c) 2012-2018 Deductive, all rights reserved
# -----------------------------------------
#  This code is licensed under MIT license (see license.txt for details)

import gzip
import logging
import multiprocessing
import os
import shutil
import unittest
from random import random
from time import sleep

import pyarrow as pa
from pandas.testing import assert_frame_equal
from s3fs import S3FileSystem

from newtools import CSVDoggo, log_to_stdout
from newtools import PandasDoggo, FileDoggo, S3Location, DoggoFileSystem, DoggoLock, DoggoWait, DynamoDogLock
from newtools.optional_imports import boto3, AWSRetry
from newtools.optional_imports import numpy as np
from newtools.optional_imports import pandas as pd
from .base_test import BaseTest

s3_bucket = "newtools-testing"

log_to_stdout("newtools", level=logging.DEBUG)

class TestDoggos(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.base_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'test_data', 'pandas_doggo_tests')

        # Upload test data to s3
        cls.s3_base_path = f"s3://{s3_bucket}/generic/"
        cls.s3 = S3FileSystem()
        for file in cls.s3.glob(f"{cls.s3_base_path}*"):
            cls.s3.rm(file)

        cls.s3.put(lpath=os.path.join(cls.base_path, "csv", "generic"),
                   rpath=cls.s3_base_path,
                   recursive=True)

    @classmethod
    def tearDownClass(cls):
        for file in cls.s3.glob(f"{cls.s3_base_path}*"):
            cls.s3.rm(file)

    def test_read_csv_local(self):
        """read a local csv"""

        expected_df = pd.DataFrame({'e-mail': {0: 'n@gmail.com', 1: 'p@gmail.com', 2: 'h@gmail.com', 3: 's@gmail.com',
                                               4: 'l@gmail.com', 5: 'v@gmail.com', 6: None},
                                    'number': {0: 0, 1: 1, 2: 0, 3: 0, 4: 1, 5: 0, 6: 0}})
        pth = os.path.join(self.base_path, 'csv', 'generic', 'email_test.csv')
        session = boto3.Session()
        fh = PandasDoggo(boto_session=session)
        df = fh.load(pth)
        pd.testing.assert_frame_equal(df, expected_df)

    def test_read_csv_local_gzip(self):
        pth = os.path.join(self.base_path, 'csv', 'gzip', 'test.csv.gz')

        fh = PandasDoggo()
        df = fh.load_csv(pth, compression='gzip')
        print(df.to_csv(index=0))

    def test_write_csv_local_gzip(self):
        pth_in = os.path.join(self.base_path, 'csv', 'gzip', 'uncompressed.csv')
        pth_out = os.path.join(self.base_path, 'csv', 'gzip', 'write.csv.gzip')
        pth_exp = os.path.join(self.base_path, 'csv', 'gzip', 'write_target.csv.gzip')
        fh = PandasDoggo()
        df = fh.load_csv(pth_in)

        fh.save_csv(df, pth_out, compression='gzip', index=0)
        with gzip.open(pth_out) as r, gzip.open(pth_exp) as exp:
            self.assertEqual(r.read().decode('utf-8'), exp.read().decode('utf-8'))

        os.remove(pth_out)

    def test_write_csv_local_gzip_infer(self):
        pth_in = os.path.join(self.base_path, 'csv', 'gzip', 'uncompressed.csv')
        pth_out = os.path.join(self.base_path, 'csv', 'gzip', 'write.csv.gzip')
        pth_exp = os.path.join(self.base_path, 'csv', 'gzip', 'write_target.csv.gzip')
        fh = PandasDoggo()
        df = fh.load_csv(pth_in)

        fh.save(df, pth_out, index=0)
        with gzip.open(pth_out) as r, gzip.open(pth_exp) as exp:
            self.assertEqual(r.read().decode('utf-8'), exp.read().decode('utf-8'))

        os.remove(pth_out)

    def test_read_csv_local_gzip_infer(self):
        pth = os.path.join(self.base_path, 'csv', 'gzip', 'test.csv.gz')

        fh = PandasDoggo()
        df = fh.load(pth)

    def test_read_csv_local_zip_infer(self):
        with self.assertRaises(NotImplementedError):
            pth = os.path.join(self.base_path, 'csv', 'gzip', 'test.csv.zip')
            fh = PandasDoggo()
            df = fh.load(pth)

    def test_write_csv_local(self):
        df = pd.DataFrame({'e-mail': {0: 'n@gmail.com', 1: 'p@gmail.com', 2: 'h@gmail.com', 3: 's@gmail.com',
                                      4: 'l@gmail.com', 5: 'v@gmail.com', 6: None},
                           'number': {0: 0, 1: 1, 2: 0, 3: 0, 4: 1, 5: 0, 6: 0}})
        pth = os.path.join(self.base_path, 'csv', 'generic', 'email_test_copy.csv')
        expected_pth = os.path.join(self.base_path, 'csv', 'generic', 'email_test.csv')
        fh = PandasDoggo()
        fh.save(df, pth, index=None)
        with FileDoggo(pth) as p, FileDoggo(expected_pth) as p2:
            self.assertEqual(p.read().strip(), p2.read().strip())

        os.remove(pth)

    def test_read_pq_local(self):
        """read a local parquet"""
        fh = PandasDoggo()
        pth = os.path.join(self.base_path, 'parquet', 'data.parquet')
        df = fh.load(pth)

    def test_read_pq_local_snappy(self):
        pth = os.path.join(self.base_path, 'parquet', 'snappy', 'compressed.snappy.parquet')

        fh = PandasDoggo()
        df = fh.load_parquet(pth)
        print(df.to_parquet(pth, index=0))

    def test_write_pq_local(self):
        fh = PandasDoggo()
        df = pd.read_csv(os.path.join(self.base_path, 'parquet', 'emails.csv'))
        fh.save_parquet(df, os.path.join(self.base_path, 'parquet', 'emails.parquet'))
        os.remove(os.path.join(self.base_path, 'parquet', 'emails.parquet'))

    def test_write_pq_snappy(self):
        pth_in = os.path.join(self.base_path, 'parquet', 'snappy', 'uncompressed.parquet')
        pth_out = os.path.join(self.base_path, 'parquet', 'snappy', 'compressed.snappy.parquet')
        pth_exp = os.path.join(self.base_path, 'parquet', 'snappy', 'writing_target.snappy.parquet')
        fh = PandasDoggo()
        df = pd.read_parquet(pth_in)
        fh.save_parquet(df, pth_out)
        r = pd.read_parquet(pth_out, engine='pyarrow')
        exp = pd.read_parquet(pth_exp, engine='pyarrow')

        pd.testing.assert_frame_equal(r, exp)

    def test_read_pq_local_with_args(self):
        fh = PandasDoggo()
        pth = os.path.join(self.base_path, 'parquet', 'data.parquet')
        df = fh.load_parquet(pth, columns=['country', 'birthdate', 'salary', 'title'])
        self.assertEqual(sorted(list(df.columns)), sorted(['country', 'birthdate', 'salary', 'title']))

    def test_pq_schema_int64(self):
        new_file_path = os.path.join(self.base_path,
                                     'parquet',
                                     'new_data.parquet')
        test_df = pd.DataFrame(
            {"col1": [1, 2, 3, 4, 5], "col2": ["a", "s", "6", "7", "t"]})
        fields = [
            pa.field("col1", pa.int64()),
            pa.field("col2", pa.string())
        ]
        my_schema = pa.schema(fields)
        fh = PandasDoggo()

        fh.save_parquet(test_df, new_file_path, schema=my_schema)

        pd.testing.assert_frame_equal(test_df, fh.load_parquet(new_file_path))
        os.remove(new_file_path)

    def test_pq_schema_int32(self):
        # Tests the down casting of int64 to int32 without explicit conversion
        new_file_path = os.path.join(self.base_path,
                                     'parquet',
                                     'new_data.parquet')
        test_df = pd.DataFrame(
            {"col1": [1, 2, 3, 4, 128], "col2": ["a", "s", "6", "7", "t"]})
        fields = [
            pa.field("col1", pa.int32()),
            pa.field("col2", pa.string())
        ]
        test_df = test_df.astype({"col1": 'Int32'})
        my_schema = pa.schema(fields)
        fh = PandasDoggo()

        fh.save_parquet(test_df, new_file_path, schema=my_schema)

        pd.testing.assert_frame_equal(test_df, fh.load_parquet(new_file_path))
        os.remove(new_file_path)

    def test_pq_schema_int32_nulls(self):
        new_file_path = os.path.join(self.base_path,
                                     'parquet',
                                     'new_data_nulls.parquet')
        test_df = pd.DataFrame(
            {"col1": pd.Series([1, 2, 3, 4, 128, pd.NA, None], dtype='Int32'),
             "col2": ["a", "s", "6", "7", "t", 'pd.NA', 'None']})
        fields = [
            pa.field("col1", pa.int32()),
            pa.field("col2", pa.string())
        ]

        my_schema = pa.schema(fields)
        fh = PandasDoggo()

        fh.save_parquet(test_df, new_file_path, schema=my_schema)

        pd.testing.assert_frame_equal(test_df, fh.load_parquet(new_file_path))
        os.remove(new_file_path)

    def test_pq_schema_int64_nulls(self):
        new_file_path = os.path.join(self.base_path,
                                     'parquet',
                                     'new_data_nulls.parquet')
        test_df = pd.DataFrame(
            {"col1": pd.Series([1, 2, 3, 4, 128, pd.NA, None], dtype='Int64'),
             "col2": ["a", "s", "6", "7", "t", 'pd.NA', 'None']})
        fields = [
            pa.field("col1", pa.int64()),
            pa.field("col2", pa.string())
        ]

        my_schema = pa.schema(fields)
        fh = PandasDoggo()

        fh.save_parquet(test_df, new_file_path, schema=my_schema)

        pd.testing.assert_frame_equal(test_df, fh.load_parquet(new_file_path))
        os.remove(new_file_path)

    def test_pq_schema_datetime(self):
        new_file_path = os.path.join(self.base_path,
                                     'parquet',
                                     'new_data.parquet')
        rng = pd.date_range('2015-02-24', periods=5, freq='23H')
        test_df = pd.DataFrame({'Date': rng, 'Val': np.random.randn(len(rng))})

        test_df['Date'] = pd.to_datetime(test_df['Date'], unit='D')
        fields = [
            pa.field("Date", pa.timestamp('ms')),
            pa.field("Val", pa.float64())
        ]
        my_schema = pa.schema(fields)
        fh = PandasDoggo()

        fh.save_parquet(test_df, new_file_path, schema=my_schema)

        pd.testing.assert_frame_equal(test_df, fh.load_parquet(new_file_path))
        os.remove(new_file_path)

    def test_read_csv_s3(self):
        """read a csv from s3"""
        expected_df = pd.DataFrame({'e-mail': {0: 'n@gmail.com', 1: 'p@gmail.com', 2: 'h@gmail.com', 3: 's@gmail.com',
                                               4: 'l@gmail.com', 5: 'v@gmail.com', 6: None},
                                    'number': {0: 0, 1: 1, 2: 0, 3: 0, 4: 1, 5: 0, 6: 0}})
        pth = S3Location('newtools-testing').join('generic', 'email_test.csv')
        session = boto3.Session()
        fh = PandasDoggo(boto_session=session)
        df = fh.load(pth)
        pd.testing.assert_frame_equal(df, expected_df)

    def test_write_csv_s3(self):
        df = pd.DataFrame({'e-mail': {0: 'n@gmail.com', 1: 'p@gmail.com', 2: 'h@gmail.com', 3: 's@gmail.com',
                                      4: 'l@gmail.com', 5: 'v@gmail.com', 6: None},
                           'number': {0: 0, 1: 1, 2: 0, 3: 0, 4: 1, 5: 0, 6: 0}})
        pth = S3Location('newtools-testing').join('generic', 'email_test_copy.csv')
        expected_pth = S3Location('newtools-testing').join('generic', 'email_test.csv')
        fh = PandasDoggo()
        fh.save(df, pth, index=None)
        with FileDoggo(pth) as p, FileDoggo(expected_pth) as p2:
            self.assertEqual(p.read().strip(), p2.read().strip())

        s3 = boto3.resource('s3')
        obj = s3.Object(pth.bucket, pth.key)
        obj.delete()

    def test_write_csv_s3_gzip(self):
        df = pd.DataFrame({'e-mail': {0: 'n@gmail.com', 1: 'p@gmail.com', 2: 'h@gmail.com', 3: 's@gmail.com',
                                      4: 'l@gmail.com', 5: 'v@gmail.com', 6: None},
                           'number': {0: 0, 1: 1, 2: 0, 3: 0, 4: 1, 5: 0, 6: 0}})
        pth = S3Location('newtools-testing').join('generic', 'email_test_copy.csv.gz')

        fh = PandasDoggo()

        fh.save(df, pth, index=None, compression="gzip")

        pd.testing.assert_frame_equal(fh.load_csv(pth), df)

        s3 = boto3.resource('s3')
        obj = s3.Object(pth.bucket, pth.key)
        obj.delete()

    def test_read_pq_s3(self):
        """read a parquet from s3"""
        fh = PandasDoggo()

        df_s3 = fh.load_parquet('s3://07092018pqtest/data.parquet')

        pth_local = os.path.join(self.base_path, 'parquet', 'data.parquet')
        df_l = fh.load_parquet(pth_local)

        pd.testing.assert_frame_equal(df_s3, df_l)

    def test_write_pq_s3(self):
        fh = PandasDoggo()
        pth_local = os.path.join(self.base_path, 'parquet', 'data.parquet')
        df_l = fh.load_parquet(pth_local)
        s3 = boto3.resource('s3')
        obj = s3.Object('07092018pqtest', 'write.parquet')
        obj.delete()  # just in case
        fh.save(df_l, 's3://07092018pqtest/write.parquet')

        obj.load()  # checks that it was written
        obj.delete()

    def test_write_pq_s3_snappy(self):
        df = pd.DataFrame({'e-mail': {0: 'n@gmail.com', 1: 'p@gmail.com', 2: 'h@gmail.com', 3: 's@gmail.com',
                                      4: 'l@gmail.com', 5: 'v@gmail.com', 6: None},
                           'number': {0: 0, 1: 1, 2: 0, 3: 0, 4: 1, 5: 0, 6: 0}})
        pth = S3Location('newtools-testing').join('generic', 'email_test_copy.parquet')

        fh = PandasDoggo()

        fh.save(df, pth, index=None, compression="snappy")

        pd.testing.assert_frame_equal(fh.load_parquet(pth), df)

        s3 = boto3.resource('s3')
        obj = s3.Object(pth.bucket, pth.key)
        obj.delete()

    def test_cant_guess_format(self):
        fh = PandasDoggo()
        with self.assertRaises(ValueError):
            fh.load('some_text_file')

    def test_invalid_format_detected(self):
        fh = PandasDoggo()
        with self.assertRaises(ValueError):
            fh.load('some_text_file.dat')
        with self.assertRaises(ValueError):
            fh.save(pd.DataFrame(), 'some_text_file.dat')

    def test_invalid_format_passed(self):
        fh = PandasDoggo()
        with self.assertRaises(ValueError):
            fh.load('some_text_file.dat', file_format='bees?')

    def test_whos_a_good_boi_den(self):
        with self.assertRaises(ValueError):
            with FileDoggo('some_path_to_a_file', mode='justusewb', is_s3=True) as f:
                f.write('should have read the docs')


class TestCSVDoggo(unittest.TestCase):
    pass


class TestDoggoFileSystem(BaseTest):

    def _test(self, root):
        dfs = DoggoFileSystem()

        testdir = dfs.join(root, str(random()))

        file_1 = dfs.join(testdir, "temp_1.txt")
        file_2 = dfs.join(testdir, "tests", "temp_2.txt")

        self.assertEqual(dfs.split(file_1), (testdir, "temp_1.txt"))

        self.assertFalse(dfs.exists(file_1))
        self.assertFalse(dfs.exists(file_2))

        with dfs.open(file_1, "wb") as f:
            f.write(b"woof")

        self.assertTrue(dfs.exists(file_1))
        self.assertFalse(dfs.exists(file_2))

        dfs.cp(file_1, file_2)
        self.assertTrue(dfs.exists(file_1))
        self.assertTrue(dfs.exists(file_2))

        paths = dfs.glob(dfs.join(testdir, "**"))
        self.assertIn(file_1, paths)
        self.assertIn(file_2, paths)

        dfs.rm(file_2)
        self.assertTrue(dfs.exists(file_1))
        self.assertFalse(dfs.exists(file_2))

        dfs.mv(file_1, file_2)
        self.assertFalse(dfs.exists(file_1))
        self.assertTrue(dfs.exists(file_2))

        self.assertEqual(dfs.size(file_2), 4)

    def test_mixed_local_s3(self):
        dfs = DoggoFileSystem()
        root = self.tempdir
        testdir = dfs.join(root, str(random()))

        file_1 = dfs.join(testdir, "temp_1.txt")
        file_2 = dfs.join(testdir, "tests", "temp_2.txt")
        dir_local_1 = testdir + '/'
        dir_local_2 = dfs.join(root, str(random())) + '/'
        dir_s3 = S3Location('newtools-testing').join('dfs')
        file_3 = dfs.join(dir_local_2, "temp_1.txt")
        file_4 = dfs.join(dir_local_2, "tests", "temp_2.txt")
        s3_loc_1 = S3Location('newtools-testing').join('dfs', 'temp_1.txt')
        s3_loc_2 = S3Location('newtools-testing').join('dfs', 'temp_2.txt')
        dfs.rm(s3_loc_1)
        dfs.rm(s3_loc_2)

        with self.assertRaises(FileNotFoundError):
            dfs.cp(s3_loc_1, s3_loc_2)

        with self.assertRaises(FileNotFoundError):
            dfs.mv(s3_loc_1, s3_loc_2)

        self.assertEqual(dfs.split(file_1), (testdir, "temp_1.txt"))

        self.assertFalse(dfs.exists(file_1))
        self.assertFalse(dfs.exists(file_2))

        with dfs.open(file_1, "wb") as f:
            f.write(b"woof")

        self.assertTrue(dfs.exists(file_1))
        self.assertFalse(dfs.exists(file_2))
        self.assertFalse(dfs.exists(file_3))
        self.assertFalse(dfs.exists(file_4))

        dfs.cp(file_1, file_2)
        self.assertTrue(dfs.exists(file_1))
        self.assertTrue(dfs.exists(file_2))
        dfs.cp(dir_local_1, dir_local_2)
        self.assertTrue(dfs.exists(file_1))
        self.assertTrue(dfs.exists(file_2))
        self.assertTrue(dfs.exists(file_3))
        self.assertTrue(dfs.exists(file_4))
        shutil.rmtree(dir_local_2)
        dfs.cp(file_1, s3_loc_1)
        self.assertTrue(dfs.exists(s3_loc_1))
        dfs.cp(s3_loc_1, s3_loc_2)
        self.assertTrue(dfs.exists(s3_loc_2))

        paths = dfs.glob(dfs.join(testdir, "**"))
        self.assertIn(file_1, paths)
        self.assertIn(file_2, paths)

        dfs.rm(file_2)
        self.assertTrue(dfs.exists(file_1))
        self.assertFalse(dfs.exists(file_2))
        dfs.cp(s3_loc_2, file_2)
        self.assertTrue(dfs.exists(file_2))
        dfs.rm(file_2)

        dfs.mv(file_1, file_2)
        self.assertFalse(dfs.exists(file_1))
        self.assertTrue(dfs.exists(file_2))
        dfs.mv(file_2, s3_loc_1)
        self.assertFalse(dfs.exists(file_2))
        self.assertTrue(dfs.exists(s3_loc_1))
        dfs.mv(s3_loc_1, s3_loc_2)
        self.assertFalse(dfs.exists(s3_loc_1))
        self.assertTrue(dfs.exists(s3_loc_2))
        dfs.mv(s3_loc_2, file_1)
        self.assertFalse(dfs.exists(s3_loc_2))
        self.assertTrue(dfs.exists(file_1))

        dfs.mv(dir_local_1, dir_local_2)
        self.assertFalse(dfs.exists(file_1))
        self.assertTrue(dfs.exists(file_3))

        dfs.mv(dir_local_2, dir_s3)
        self.assertFalse(dfs.exists(file_3))
        self.assertTrue(dfs.exists(s3_loc_1))

        dfs.rm(s3_loc_1)
        dfs.rm(s3_loc_2)

    def test_s3(self):
        self._test(S3Location("s3://07092018pqtest/fs/"))

        dfs = DoggoFileSystem()
        self.assertEqual(dfs.split("s3://07092018pqtest/temp.txt"), ("s3://07092018pqtest/", "temp.txt"))

    def test_file(self):
        self._test(self.tempdir)


class TestDoggoWait(unittest.TestCase):
    def test_basic(self):
        dw = DoggoWait(wait_period=1, time_out_seconds=2)

        with self.assertRaises(ValueError):
            dw.timed_out()

        dw.start_timeout()
        self.assertEqual(dw.timed_out(), False)
        dw.check_timeout()
        dw.wait()
        with self.assertRaises(TimeoutError):
            dw.check_timeout()


def check_lock(file):
    test_str = str(random())
    with DoggoLock(file, wait_period=0.1, time_out_seconds=20, maximum_age=30):
        with FileDoggo(file, "wb") as f:
            f.write(test_str.encode('UTF-8'))

        with FileDoggo(file, "rb") as f:
            assert f.read().decode('UTF-8') == test_str

        sleep(0.1)

    return True

class BaseLockTest(BaseTest):
    LockClass = None

    def test_lock_timeout(self):
        if self.LockClass is not None:
            file = os.path.join(self.tempdir, f'test_lock_timeout.{random()}tmp')
            with self.LockClass(file, wait_period=1, time_out_seconds=5, maximum_age=20):
                sleep(0.1)
                with self.assertRaises(TimeoutError):
                    with self.LockClass(file, wait_period=1, time_out_seconds=2, maximum_age=20):
                        pass

    def test_basic(self):
        if self.LockClass is not None:
            file = os.path.join(self.tempdir, f'test_basic.{random()}.tmp')
            check_lock(file)
            check_lock(file)

    def test_lock_ages(self):
        if self.LockClass is not None:
            file = os.path.join(self.tempdir, f'test_lock_ages.{random()}.tmp')
            lock_1 = self.LockClass(file, wait_period=1, time_out_seconds=2, maximum_age=10)
            lock_2 = self.LockClass(file, wait_period=1, time_out_seconds=2, maximum_age=10)

            lock_1.acquire()
            sleep(0.1)

            with self.assertRaises(TimeoutError):
                lock_2.acquire()

            sleep(10)

            lock_2.acquire()

            lock_1.release()
            lock_2.release()


class TestDoggoLockS3(BaseLockTest):
    tempdir = 's3://newtools-tests-data/locks/'
    LockClass = DoggoLock


class TestDoggoLockLocal(BaseLockTest):
    LockClass = DoggoLock

    def test_lock_multiprocess(self):
        if self.LockClass is not None:
            file = os.path.join(self.tempdir, f'test_lock_multiprocess.{random()}.tmp')

            processes=list()
            pool = multiprocessing.Pool()
            for _ in range(0, min(10, multiprocessing.cpu_count() -1)):
                processes.append(
                    pool.apply_async(check_lock,
                                     (file,))
                )

            pool.close()
            pool.join()

            self.assertTrue(all([x.get() for x in processes]))


class TestDynamoDogLock(BaseLockTest):
    LockClass = DynamoDogLock

    def test_lock_create_dynamo_table(self):
        boto3.resource("dynamodb", region_name="us-east-1").Table(DynamoDogLock._table_name).delete()
        waiter = boto3.client("dynamodb", region_name="us-east-1").get_waiter('table_not_exists')
        waiter.wait(TableName=DynamoDogLock._table_name,
                    WaiterConfig={'Delay': 2, 'MaxAttempts': 120})

        file = os.path.join(self.tempdir, 'test_lock_ages.tmp')
        with self.LockClass(file, wait_period=1, time_out_seconds=2, maximum_age=5):
            pass


class TestPartitionDoggo(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.df = pd.DataFrame({"col1": [1, 2, 3, 4, 5, 1, "4", 3, 4, "4", "4", 4, 2, 3, "4"],
                               "col2": [6, 7, 8, 9, 10, 5, 6, 7, 5, "10", 7, 6, 8, 7, 6],
                               "col3": [11, 12, 13, 14, 15, "a", "b", "y", 1, 3, 15, 66, 77, 13, 16],
                               "col4": ["sda", "", 16, "17", 18, "14", 12, 13, 14, 5, 6, 45, 34, 76, 87]}).astype(
            {"col4": str, "col2": int, "col3": str})
        try:
            os.mkdir(os.path.join(os.path.dirname(__file__), "test_data/athena_partition/output"))
        except FileExistsError:
            pass
        cls.base_path = os.path.join(os.path.dirname(__file__), "test_data/athena_partition/output")

    def test_csv_out(self):
        expected_data = os.path.join(os.path.dirname(__file__), "test_data/athena_partition/expected_csv.csv")
        for file in os.listdir(self.base_path):
            if not file.startswith('.'):
                file_path = os.path.join(self.base_path, file)
                shutil.rmtree(file_path)

        partition_cat = ["col1"]
        cols_keep = ["col2", "col4"]
        output_data_path = os.path.join(self.base_path, "col1=4")
        PandasDoggo().save_partitioned(df=self.df[partition_cat + cols_keep].astype(dict(zip(partition_cat, [int]))),
                                       partition_columns=partition_cat,
                                       base_path=self.base_path,
                                       name="/test_name_",
                                       suffix=".csv",
                                       index=False)
        self.assertEqual(1, len([name for name in os.listdir(output_data_path)]))
        csv = CSVDoggo()
        expected = csv.load_df(expected_data)
        output = csv.load_df(os.path.join(output_data_path, *os.listdir(output_data_path)))

        assert_frame_equal(expected, output, check_like=True)

        for file in os.listdir(self.base_path):
            if not file.startswith('.'):
                file_path = os.path.join(self.base_path, file)
                shutil.rmtree(file_path)

    def test_pq_out(self):
        expected_data = os.path.join(os.path.dirname(__file__), "test_data/athena_partition/expected_parquet.parquet")
        for file in os.listdir(self.base_path):
            if not file.startswith('.'):
                file_path = os.path.join(self.base_path, file)
                shutil.rmtree(file_path)

        partition_cat = ["col1"]
        cols_keep = ["col2", "col4"]
        output_data_path = os.path.join(self.base_path, "col1=4")
        PandasDoggo().save_partitioned(df=self.df[partition_cat + cols_keep].astype(dict(zip(partition_cat, [int]))),
                                       partition_columns=partition_cat,
                                       base_path=self.base_path,
                                       name="/test_name_",
                                       suffix=".parquet",
                                       index=False)

        dog = PandasDoggo()
        expected = dog.load_parquet(expected_data)
        output = dog.load_parquet(os.path.join(output_data_path, *os.listdir(output_data_path)))
        assert_frame_equal(expected, output, check_like=True)

        for file in os.listdir(self.base_path):
            if not file.startswith('.'):
                file_path = os.path.join(self.base_path, file)
                shutil.rmtree(file_path)

    def test_multi_part(self):
        expected_data = os.path.join(os.path.dirname(__file__), "test_data/athena_partition/expected_multi_csv.csv")
        for file in os.listdir(self.base_path):
            if not file.startswith('.'):
                file_path = os.path.join(self.base_path, file)
                shutil.rmtree(file_path)

        csv = CSVDoggo(base_path=self.base_path)
        partition_cat = ["col1", "col2"]
        output_data_path = os.path.join(self.base_path, "col1=4/col2=6")
        PandasDoggo().save_partitioned(df=self.df.astype(dict(zip(partition_cat, [int, int]))),
                                       partition_columns=partition_cat,
                                       base_path=self.base_path,
                                       name="/test_name_",
                                       suffix=".parquet",
                                       index=False)

        self.assertEqual(5, len([name for name in os.listdir(os.path.join(self.base_path, "col1=4"))]))
        dog = PandasDoggo()
        expected = dog.load(expected_data, dtype=object)
        output = dog.load(os.path.join(output_data_path, *os.listdir(output_data_path)))

        assert_frame_equal(expected, output, check_like=True)

        for file in os.listdir(self.base_path):
            if not file.startswith('.'):
                file_path = os.path.join(self.base_path, file)
                shutil.rmtree(file_path)

    def test_expected_errors(self):

        with self.assertRaises(ValueError):
            PandasDoggo().save_partitioned(df=self.df.astype(dict(zip(["col4"], [int, int]))),
                                           partition_columns=["col4"],
                                           base_path=self.base_path,
                                           name="/test_name_",
                                           suffix=".parquet",
                                           index=False)

        with self.assertRaises(KeyError):
            PandasDoggo().save_partitioned(df=self.df.astype(dict(zip(["col1", "col7"], [int, int]))),
                                           partition_columns=["col1"],
                                           base_path=self.base_path,
                                           name="/test_name_",
                                           suffix=".parquet",
                                           index=False)

        with self.assertRaises(ValueError):
            PandasDoggo().save_partitioned(df=self.df.astype(dict(zip(["col3"], [int, int]))),
                                           partition_columns=["col3"],
                                           base_path=self.base_path,
                                           name="/test_name_",
                                           suffix=".parquet",
                                           index=False)



from .doggo import PandasDoggo, FileDoggo
from .fs import DoggoFileSystem
from .lock import DoggoLock, DoggoWait, DynamoDogLock
from .csv import CSVDoggo

__all__ = ['CSVDoggo', 'PandasDoggo','DoggoFileSystem', 'DoggoLock', 'DoggoWait', 'DynamoDogLock']


# (c) 2012-2020 Deductive, all rights reserved
# -----------------------------------------
#  This code is licensed under MIT license (see license.txt for details)


import gzip
import json
import logging
import os
import re
from datetime import datetime
from io import BytesIO

from newtools.aws import S3Location
from newtools.optional_imports import AWSRetry
from newtools.optional_imports import boto3
from newtools.optional_imports import pandas as pd


def _gzip_wrapped(func):
    def gzip_open(self, *args, **kwargs):
        if self.compression == 'gzip':
            f = func(self, *args, **kwargs)
            self._to_close.append(f)
            return gzip.open(f, self.mode)
        return func(self, *args, **kwargs)

    return gzip_open


class PandasDoggo:
    """
    Is a Panda a doggo?

    .. parsed-literal::

        ░░░░░░░░▄██▄░░░░░░▄▄░░
        ░░░░░░░▐███▀░░░░░▄███▌
        ░░▄▀░░▄█▀▀░░░░░░░░▀██░
        ░█░░░██░░░░░░░░░░░░░░░
        █▌░░▐██░░▄██▌░░▄▄▄░░░▄
        ██░░▐██▄░▀█▀░░░▀██░░▐▌
        ██▄░▐███▄▄░░▄▄▄░▀▀░▄██
        ▐███▄██████▄░▀░▄█████▌
        ▐████████████▀▀██████░
        ░▐████▀██████░░█████░░
        ░░░▀▀▀░░█████▌░████▀░░
        ░░░░░░░░░▀▀███░▀▀▀░░░░


    A class designed to simplify file IO operation to and from local or s3 files, with specific functionality for csv and parquet file formats

    Key features
    ------------

    - read / write csv and parquet files
    - read and write both locally and to s3
    - support gzip, snappy and zip compression
    - sensible s3 handling of different profiles / credentials

    Usage
    -----

    .. code-block:: python

        from newtools import PandasDoggo

        fh = PandasDoggo()

        df = fh.load('filename.ext')

        fh.save(df, 'path/to/new_file/ext')

    """

    def __init__(self,
                 boto_session=None,
                 logger=logging.getLogger("newtools.pandas_doggo")):
        """

        :param boto_session: optional boto3 session to use for
        :type boto_session: boto3.Session
        :param logger: the logger to use
        """

        if boto_session:
            self.s3_client = boto_session.client('s3')
        else:
            self.s3_client = None

        self._logger = logger

    @staticmethod
    def _extract_file_format(path):

        for part in reversed(path.split(".")[-2:]):
            if part in ('csv', 'parquet', 'pq'):
                return part

        raise ValueError(f'could not determine format of path: {path}, use file_format param to specify')

    @staticmethod
    def _extract_compression(path):
        # extend for other compression

        if path.endswith('.gzip') or path.endswith('.gz'):
            return 'gzip'
        if ('.snappy' in path) or path.endswith('.sz'):
            return 'snappy'
        if path.endswith('.zip'):
            return 'zip'

    def load(self, path, file_format=None, compression=None, *args, **kwargs):
        """Load a file into a Pandas.DataFrame from local or s3 locations.

        :param path: required. Can be s3 or local s3 must be in `s3://` format - accepts `S3Location`
        :type path: str
        :param file_format: None. Autodetects from path, can be set to `csv` or `parquet` to explicitly force a format
        :param compression: optional, 'gzip', 'snappy' or None. Autodetects from path if path ends in `.gz`, `.gzip` or contains `.snappy`
        :param args: args to pass to the panda to load the file
        :param kwargs: kwargs to pass to the panda to load the file eg columns=['subset', 'of', 'columns']
        :return: Pandas.DataFrame
        """

        file_format = file_format or self._extract_file_format(path)

        if file_format not in ('csv', 'parquet', 'pq'):
            raise ValueError(f'detected format:{format} is not recognized file format for load')

        if file_format == 'csv':
            return self.load_csv(path, compression=compression, *args, **kwargs)

        elif file_format in ('parquet', 'pq'):
            return self.load_parquet(path, compression=compression, *args, **kwargs)

    def load_csv(self, path, compression=None, *args, **kwargs):
        """alias for .load(path, format='csv')"""
        if 'chunksize' in kwargs:
            raise NotImplementedError("PandasDoggo does not support chunksize for loading data frames")

        compression = compression or self._extract_compression(path)
        fm = FileDoggo(path, mode='rb', client=self.s3_client, compression=compression)
        with fm as f:
            df = pd.read_csv(f, *args, **kwargs)

        return df

    def load_parquet(self, path, compression=None, *args, **kwargs):
        """alias for .load(path, format='parquet')"""

        compression = compression or self._extract_compression(path)
        fm = FileDoggo(path, mode='rb', client=self.s3_client, compression=compression)
        with fm as f:
            df = pd.read_parquet(f, engine='pyarrow', *args, **kwargs)

        return df

    def save(self, df, path, file_format=None, compression=None, *args, **kwargs):
        """Save a file into a Pandas.DataFrame from local or s3 locations.

        :param df: Data frame
        :type df: Pandas.DataFrame
        :param path: required. Can be s3 or local s3 must be in `s3://` format - accepts `S3Location`
        :type path: str
        :param file_format: 'None. Autodetects from path, can be set to `csv` or `parquet` to explicitly force a format
        :param compression: None. Supports gzip and snappy, autodetects from path if path ends in `.gz`, `.gzip` or contains `.snappy`
        :param args: args to pass to the panda to save the file
        :param kwargs: kwargs to pass to the panda to save the file eg index=None
        """

        file_format = file_format or self._extract_file_format(path)

        if file_format == 'csv':
            return self.save_csv(df, path, compression, *args, **kwargs)

        elif file_format in ('parquet', 'pq'):
            return self.save_parquet(df, path, compression, *args, **kwargs)

    def save_csv(self, df, path, compression=None, *args, **kwargs):
        """Alias for .save(df, format='csv')"""
        compression = compression or self._extract_compression(path)
        fm = FileDoggo(path, mode='wb', client=self.s3_client, compression=compression)
        with fm as f:
            f.write(df.to_csv(None, *args, **kwargs).encode('utf-8'))

    def save_parquet(self, df, path, compression=None, *args, **kwargs):
        """Alias for .save(df, format='parquet')"""

        compression = compression or self._extract_compression(path)
        fm = FileDoggo(path, mode='wb', client=self.s3_client, compression=compression)
        with fm as f:
            df.to_parquet(f, *args, **kwargs)

    def save_partitioned(self,
                         df,
                         base_path,
                         name,
                         suffix,
                         partition_columns,
                         partition_string="",
                         date_time_format="%Y-%m-%d_%H%M%S.%f",
                         compression=None, *args, **kwargs):
        """Save a data frame into multiple files, partitioned by the specified columns. The base path can be
        local file system, or S3.

        Based on athena_partition from old tools. Notable differences are as follows:
        * PandasDoggo saves the index as default so pass index=False for comparable behavior
        * base_path - new parameter that was previously taken from the passed file_handler
        * partition_columns - replaces partition_categories
        * partition_dtypes - not supported so please apply any changes to dtype before passing using df.astype(dict(zip(["col1", "col7"], [int, int])
        * columns_to_keep - not supported, please only send a slice of the data frame with the columns you want to partition or save
        * file_handler - not supported, uses the PandasDoggo class.

        :param df: The data frame to be partitioned
        :param base_path: The base path to save the files to
        :param name: If provided all files filename will start with this
        :param suffix: The extension the file should be saved with, .csv for csv, and .parquet for parquet
        :param partition_columns: The columns to partition the data on
        :param partition_string: Allows formatting folder names, will be dependant on how many partition categories there
                are, defaults to creating hive-format folders and sub folders in order of partitioning
        :param date_time_format: To minimise chances of overwrite the saved files contain the date time of when this
                function was called, this param specifies the format of the date time
        :param compression: None. Supports gzip and snappy, autodetects from suffix if it ends in `.gz`, `.gzip` or contains `.snappy`
        :param args: args to pass to the panda to save the file
        :param kwargs: kwargs to pass to the panda to save the file eg index=None
        :return: Returns a full list of all file paths created, doesnt return base path as part of this
        """

        if any(df[partition_columns].isna().any()) or any((df[partition_columns] == "").any()):  # pragma: no cover
            raise ValueError('The partition columns contain NaN values')

        # Set up the default partition string
        if not partition_string:
            partition_string = "/".join([col + "={}" for col in partition_columns])

        paths = list()

        # Create each partition file
        for partition_values, group_df in df.groupby(partition_columns, sort=False):

            # Format the partitions
            if len(partition_columns) == 1:
                partition_values = [partition_values]
            partitions = partition_string.format(*partition_values)

            # Get the time
            time = datetime.now().strftime(date_time_format)
            self._logger.info("time is {}".format(time))

            # Get the file
            file = f"{base_path}/{partitions}/{name}{datetime.now().strftime(date_time_format)}{suffix}"
            self._logger.info(json.dumps({"message": "Saving parquet to location {}".format(file)}))

            # Create the path if it doesn't exist...
            if file.startswith("/"):
                try:
                    os.makedirs(os.path.dirname(file))
                except FileExistsError:  # pragma: no cover
                    pass

            # Save the dataframe
            self.save(group_df[[col for col in df.columns if col not in partition_columns]],
                      file, compression, *args, **kwargs)

            # Add to the list to return
            paths.append(file)

        return paths


class FileDoggo:
    """

He fetches the things you want. Tv remotes, slippers, newspapers, files. Mostly files though.

.. code-block:: python

    from newtools import FileDoggo
    import boto3

    path = 's3 or local path'
    s3_client = boto3.Session().client('s3')
    with FileDoggo(path, mode='rb', client=s3_client) as f:
        f.read()

This is written to treat local paths and s3 paths the same, returning a file like object for either.

"""
    _connection = None

    def __init__(self, path, mode='rb', is_s3=None, client=None, compression=None):
        """
        Creates the class

        :param path: path to file to connect to. s3 or local path
        :type path: str or S3Location
        :param mode: opening mode for file, note s3 only allows 'rb' or 'wb'
        :param is_s3: force Doggo to treat path as an S3Location, otherwise autodetects if beings with "s3://"
        :type is_s3: bool
        :param client: optional `boto3.Session().Client('s3')` instance to use for s3 operations
        :param compression: None, 'gzip' or 'snappy'.
        """

        self.buffer = BytesIO()
        self.mode = mode

        self.is_s3 = is_s3 or path.startswith('s3://')
        if self.is_s3:
            # allows you to specify s3 for paths that don't begin with s3:// for whatever reason you might want
            self.path = S3Location(path)
        else:
            self.path = path

        self._client = client
        if compression in [None, 'gzip', 'snappy']:
            self.compression = compression
        else:
            raise NotImplementedError(f"Compression {compression} is not supported in FileDoggo")
        self._to_close = []

    @property
    def connection(self):
        if not self._connection:
            self._connection = self._connect()
        return self._connection

    @property
    def client(self):
        if not self._client:
            self._client = boto3.Session().client('s3')
        return self._client

    def close(self):

        if self._connection:
            for c in [self._connection] + self._to_close:
                if not c.closed:
                    c.close()

    def __enter__(self):
        return self.connection

    def __exit__(self, etype, value, traceback):
        # ¯\_(ツ)_/¯
        if self.is_s3 and 'w' in self.mode:
            self._write_s3()

        self.close()

    @_gzip_wrapped
    def _connect(self):

        if self.is_s3:
            if self.mode == 'rb':
                return self._read_s3()
            elif self.mode == 'wb':
                # check s3 exists? meh, writing to a buffer is so fast i think the overhead isn't worth
                return self.buffer  # write to this buffer and we'll upload later
            else:
                raise ValueError(f'mode {self.mode} is not supported')

        else:

            return open(self.path, self.mode)

    @AWSRetry.backoff(tries=5, delay=1, backoff=1, added_exceptions=['404'])
    def _read_s3(self):

        self.client.download_fileobj(Bucket=self.path.bucket, Key=self.path.key, Fileobj=self.buffer)
        self.buffer.seek(0)
        return self.buffer

    def _write_s3(self):

        if self.compression == 'gzip':
            self.connection.close()  # flushes writing to gzip

        self.buffer.seek(0)
        self.client.upload_fileobj(Bucket=self.path.bucket, Key=self.path.key, Fileobj=self.buffer)



import logging
import sys

###########################################
#                                         #
#             HERE BE DRAGONS             #
#                                         #
#                                         #
#     .     _///_,                        #
#   .      / ` ' '>                       #
#     )   o'  __/_'>                      #
#    (   /  _/  )_\'>                     #
#     ' "__/   /_/\_>                     #
#         ____/_/_/_/                     #
#        /,---, _/ /                      #
#       ""  /_/_/_/                       #
#          /_(_(_(_                 \     #
#         (   \_\_\\_               )\    #
#          \'__\_\_\_\__            ).\   #
#          //____|___\__)           )_/   #
#          |  _  \'___'_(           /'    #
#           \_ (-'\'___'_\      __,'_'    #
#           __) \  \\___(_   __/.__,'     #
#        ,((,-,__\  '", __\_/. __,'       #
#                     '"./_._._-'         #
#                                         #
# THIS FILE NEEDS TO HAVE AUTOMATED TESTS #
#                                         #
###########################################

handler = logging.StreamHandler(sys.stdout)
handler.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))


def clean_children(logger):  # pragma: nocover
    """
    Check the child handlers and remove our handler if it is set
    :param logger: the logger to check the children of
    """
    for l in [logging.getLogger(name) for name in logging.root.manager.loggerDict]:
        if l.name.startswith(logger.name):
            for h in l.handlers:
                if h == handler:
                    l.removeHandler(h)


def already_set(logger):  # pragma: nocover
    """
    Check all parent handlers to see if the handler is already set in the hierarchy
    :param logger: the logger to check the parents of
    :return: True if our handler is already set
    """
    for l in [logging.getLogger(name) for name in logging.root.manager.loggerDict]:
        if logger.name.startswith(l.name):
            for h in l.handlers:
                if h == handler:
                    return True

handler = logging.StreamHandler(sys.stdout)
handler.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))


def clean_children(logger):  # pragma: nocover
    """
    Check the child handlers and remove our handler if it is set
    :param logger: the logger to check the children of
    """
    for l in [logging.getLogger(name) for name in logging.root.manager.loggerDict]:
        if l.name.startswith(logger.name):
            for h in l.handlers:
                if h == handler:
                    l.removeHandler(h)


def already_set(logger):  # pragma: nocover
    """
    Check all parent handlers to see if the handler is already set in the hierarchy
    :param logger: the logger to check the parents of
    :return: True if our handler is already set
    """
    for l in [logging.getLogger(name) for name in logging.root.manager.loggerDict]:
        if logger.name.startswith(l.name):
            for h in l.handlers:
                if h == handler:
                    return True


def log_to_stdout(logger_name="deductive", level=logging.INFO):  # pragma: nocover
    """
    Adds a handler to the specified logger to pipe it to stdout.

    Checks the entire logging hierarchy for other instances of the same logger to prevent duplicates.
    :param logger_name: the logger to pip to stdout, defaults to deductive
    :param level: the level of logs to pipe to stdout, default to INFO
    :return: the logger
    """
    logger = logging.getLogger(logger_name)
    logger.setLevel(level)

    clean_children(logger)

    if not already_set(logger):
        logger.addHandler(handler)

    return logger
